import os
import sys
from typing import Dict

import graphviz
import networkx as nx
from airflow.models import DAG, BaseOperator, DagBag

from karadoc.airflow.conf import (
    get_airflow_dag_folder_location,
    get_airflow_plugin_folder_location,
)
from karadoc.airflow.utils import is_external_sensor, is_trigger


class DagNotFoundException(Exception):
    def __init__(self, dag_id: str) -> None:
        super().__init__(f"Dag not found : {dag_id}")
        self.dag_id = dag_id


class DagLoadingException(Exception):
    pass


def variable_get_mock(*args, **kwargs):
    return ""


def __delete_airflow_modules() -> None:
    """Delete all airflow modules to force Airflow to reload the environment variables AIRFLOW__CORE__PLUGINS_FOLDER
    and AIRFLOW__CORE__DAGS_FOLDER when we update them.
    """
    template_package_name = "airflow"
    template_modules = [m for m in sys.modules if m.startswith(template_package_name)]
    for m in template_modules:
        del sys.modules[m]


def build_dag_bag() -> DagBag:
    __delete_airflow_modules()
    plugins_folder = get_airflow_plugin_folder_location()
    dag_folder = get_airflow_dag_folder_location()
    os.environ["AIRFLOW__CORE__PLUGINS_FOLDER"] = plugins_folder
    os.environ["AIRFLOW__CORE__DAGS_FOLDER"] = dag_folder
    return DagBag(dag_folder=dag_folder, include_examples=False)


def get_orphan_tasks(dag: DAG):
    return [task for task in dag.tasks if len(task.upstream_list) == 0]


def task_full_name(task: BaseOperator):
    return task.dag_id + "." + task.task_id


def _get_dag(dag_bag: DagBag, dag_id: str) -> DAG:
    if dag_id not in dag_bag.dags:
        raise DagNotFoundException(dag_id)
    trigger_dag = dag_bag.dags[dag_id]
    return trigger_dag


def build_task_dependency_graph(graph: nx.DiGraph, dag: DAG, dag_bag: DagBag) -> nx.DiGraph:
    """Builds the dependency graph of all airflow tasks.
    It is able to display cross-dags dependencies by analyzing tasks of type CustomTriggerDagRunOperator
    and ExternalTaskSensor.

    :param graph:
    :param dag:
    :param dag_bag:
    :return:
    """
    for task in dag.tasks:
        try:
            graph.add_node(task_full_name(task))
            for successor in task.downstream_list:
                graph.add_edge(task_full_name(task), task_full_name(successor))
            if is_trigger(task):
                trigger_dag = _get_dag(dag_bag, task.trigger_dag_id)
                for starting_task in get_orphan_tasks(trigger_dag):
                    graph.add_edge(task_full_name(task), task_full_name(starting_task))
            if is_external_sensor(task):
                external_dag = _get_dag(dag_bag, task.external_dag_id)
                external_task = external_dag.get_task(task.external_task_id)
                graph.add_edge(task_full_name(external_task), task_full_name(task))
        except Exception as e:
            raise DagLoadingException(f"Failed to load task {task_full_name(task)} :\n" + e.args[0]) from e
    return graph


def build_graph(dag_bag: DagBag) -> nx.DiGraph:
    """Generate the full dependency graph of all airflow dags
    :return: a directed graph of full task ids ("dag_id.task_id")
    """
    graph = nx.DiGraph()  # type: nx.DiGraph
    for dag in dag_bag.dags.values():
        graph = build_task_dependency_graph(graph, dag, dag_bag)
    return graph


def transitive_closure(graph) -> nx.DiGraph:
    """Returns the transitive closure of a directed graph

    The transitive closure of G = (V,E) is a graph G+ = (V,E+) such that
    for all v,w in V there is an edge (v,w) in E+ if and only if there
    is a non-null path from v to w in G.

    Parameters
    ----------
    G : NetworkX DiGraph
        A directed graph

    Returns
    -------
    NetworkX DiGraph
        The transitive closure of `G`

    Raises
    ------
    NetworkXNotImplemented
        If `G` is not directed

    References
    ----------
    .. [1] http://www.ics.uci.edu/~eppstein/PADS/PartialOrder.py

    """
    return nx.transitive_closure(graph)


def render_graph(graph: nx.DiGraph):
    """Generates a Digraph.gv file and display it using the Graphviz library"""
    dot = graphviz.Digraph(format="png")
    dot.attr(rankdir="LR")

    dag_subgraphs: Dict[str, graphviz.Digraph] = {}

    def get_dag_subgraph(dag_id: str) -> graphviz.Digraph:
        sub = dag_subgraphs.get(dag_id)
        if sub is None:
            sub = graphviz.Digraph(name="cluster_" + dag_id)
            sub.attr(label=dag_id)
            sub.attr(style="dotted")
            dag_subgraphs[dag_id] = sub
        return sub

    for node in graph.nodes():
        dag_id, task_id = node.split(".", maxsplit=1)
        sub = get_dag_subgraph(dag_id)
        shape = "box"
        fillcolor = "aliceblue"
        label = task_id
        style = "filled"
        color = None
        sub.node(node, label=label, shape=shape, fillcolor=fillcolor, style=style, color=color)
    for (source, dest) in graph.edges():
        dot.edge(source, dest)

    for (schema, sub) in dag_subgraphs.items():
        dot.subgraph(sub)

    dot.render(view=True)


def build_dag_graph(task_graph: nx.DiGraph) -> nx.DiGraph:
    dag_graph = nx.DiGraph()  # type: nx.DiGraph
    for node in task_graph.nodes:
        node_dag = node.split(".", maxsplit=1)[0]
        dag_graph.add_node(node_dag)
    for source, dest in task_graph.edges:
        source_dag = source.split(".", maxsplit=1)[0]
        dest_dag = dest.split(".", maxsplit=1)[0]
        dag_graph.add_edge(source_dag, dest_dag)
    return dag_graph


def render_dag_graph(graph: nx.DiGraph):
    """Generates a Digraph.gv file and display it using the Graphviz library"""
    dot = graphviz.Digraph(format="png")
    dot.attr(rankdir="LR")

    nodes = {node for node in graph.nodes()}
    edges = {(source, dest) for source, dest in graph.edges() if source != dest}
    for node in nodes:
        shape = "ellipse"
        fillcolor = "aliceblue"
        label = node
        style = "filled"
        color = None
        dot.node(node, label=label, shape=shape, fillcolor=fillcolor, style=style, color=color)
    for (source, dest) in edges:
        dot.edge(source, dest)

    dot.render(view=True)



from typing import Dict, List, Optional

from karadoc.common import table_utils
from karadoc.common.model import file_index
from karadoc.common.run import load_populate


class Populate:
    input_tables: List[str] = []
    disabled: bool = False
    partitions: List[str] = []
    external_inputs: Dict = {}
    external_outputs: Dict = {}
    vars: Dict[str, str] = {}
    hide_inputs: bool = False

    def __init__(self, full_table_name):
        job = load_populate(full_table_name)
        self.input_tables = [table_utils.remove_partition_from_name(table) for table in job.list_input_tables()]
        self.disabled = job.disable
        self.partitions = job.output_partition_names
        self.external_inputs = job.external_inputs
        self.external_outputs = job.external_outputs
        self.vars = job.vars
        self.hide_inputs = job.hide_inputs


class Table:
    def __init__(self, full_name):
        self.full_name: str = full_name
        self.populate: Optional[Populate] = None
        if table_utils.populate_exists(full_name):
            self.populate = Populate(full_name)


def build_table_index() -> Dict[str, Table]:
    """Returns a dict (full_table_name -> table information) of all tables declared in Karadoc's model"""
    res = {}
    for (schema, table, path) in file_index.list_schema_table_folders():
        full_table_name = "%s.%s" % (schema, table)
        res[full_table_name] = Table(full_table_name)
    return res



from typing import Mapping, Optional

from karadoc.common.observability import LogEvent


class ConsoleEvent(LogEvent):
    """Special type of LogEvent that gets printed on stdout for user-friendly feedback"""

    extra_console_message: Optional[str] = None

    def __init__(
        self,
        message: str,
        event_type: str,
        extra: Optional[Mapping[str, str]] = None,
        extra_console_message: Optional[str] = None,
    ) -> None:
        super().__init__(message, event_type, extra)
        self.extra_console_message = extra_console_message



from typing import Dict, List, Set, Union

from pyspark import Row
from pyspark.sql import DataFrame

from karadoc.common import spark_utils
from karadoc.spark.utils import get_spark_session


def _compare_list_ignore_order(expected: Set, actual: List) -> bool:
    if len(expected) != len(actual):
        return False
    for x in expected:
        if x not in actual:
            return False
    return True


def _compare_dicts(expected: Dict, actual: Dict) -> bool:
    if len(expected) != len(actual):
        return False
    for k, expected_v in expected.items():
        if k not in actual:
            return False
        actual_v = actual[k]
        if isinstance(expected_v, Set):
            if not _compare_list_ignore_order(expected_v, actual_v):
                return False
        elif isinstance(actual_v, Set):
            if not _compare_list_ignore_order(actual_v, expected_v):
                return False
        elif expected_v != actual_v:
            return False
    return True


class MockRow:
    """This class can be used in unit tests to be easily compared with a real Spark Row

    Ignoring order for ARRAY comparison
    ----------------------------------
    When testing DataFrames with columns of type ARRAY, one might want to make their test cases ignore the ordering
    of the arrays.

    This can be done be declaring values as sets instead of lists.
    List values will check that the order matches while set values will ignore ordering

    Warning: two MockRows cannot be compared together.
    This is to prevent sets like `{MockRow(a=1), MockRow(a=1)}` to be reduced to `{MockRow(a=1)}`.

    Example:

    >>> Row(a=[1, 2, 3]) == MockRow(a=[1, 2, 3])
    True
    >>> Row(a=[3, 2, 1]) == MockRow(a=[1, 2, 3])
    False

    >>> Row(a=[1, 2, 3]) == MockRow(a={1, 2, 3})
    True
    >>> Row(a=[3, 2, 1]) == MockRow(a={1, 2, 3})
    True

    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __hash__(self):
        return hash((tuple(self.kwargs.keys()), self.kwargs.values()))

    def __eq__(self, other):
        if isinstance(other, Row):
            return _compare_dicts(self.kwargs, other.asDict())
        else:
            return NotImplemented

    def __repr__(self):
        return f"MockRow({self.kwargs})"

    def to_row(self) -> Row:
        """Transforms this MockRow into a regular pyspark.sql.Row

        >>> MockRow(a=[1, 2, 3]).to_row()
        Row(a=[1, 2, 3])
        >>> MockRow(a=[1, 2, 3]).to_row()
        Row(a=[1, 2, 3])

        >>> MockRow(a={1, 2, 3}).to_row()
        Row(a=[1, 2, 3])
        >>> MockRow(a={1, 2, 3}).to_row()
        Row(a=[1, 2, 3])

        """

        def to_row(r):
            if isinstance(r, MockRow):
                return r.to_row()
            if isinstance(r, list) or isinstance(r, set):
                return [to_row(x) for x in r]
            else:
                return r

        return Row(**{k: to_row(v) for k, v in self.kwargs.items()})


class MockDataFrame:
    """This class can be used in unit tests to be easily compared with a real Spark DataFrame.

    Ignoring row ordering
    ---------------------
    When comparing DataFrames with multiple rows, the order of rows can be ignored by declaring a set of rows instead
    of a list.


    Example:

    >>> spark = spark_utils.get_spark_session("doctest")
    >>> spark.createDataFrame([Row(a=1), Row(a=2)]) == MockDataFrame([MockRow(a=1), MockRow(a=2)])
    True
    >>> spark.createDataFrame([Row(a=1), Row(a=2)]) == MockDataFrame([MockRow(a=2), MockRow(a=1)])
    False
    >>> spark.createDataFrame([Row(a=1), Row(a=2)]) == MockDataFrame({MockRow(a=1), MockRow(a=2)})
    True
    >>> spark.createDataFrame([Row(a=1), Row(a=2)]) == MockDataFrame({MockRow(a=2), MockRow(a=1)})
    True
    >>> spark.createDataFrame([Row(a=1), Row(a=1)]) == MockDataFrame({MockRow(a=1), MockRow(a=1)})
    True

    """

    def __init__(self, rows: Union[Set[MockRow], List[MockRow]]):
        self.rows = rows

    def __eq__(self, other):
        if isinstance(other, DataFrame):
            other = other.select(spark_utils.quote_columns(sorted(other.columns)))
            other_rows = other.collect()
            if isinstance(self.rows, Set):
                return _compare_list_ignore_order(self.rows, other_rows)
            else:
                return other_rows == self.rows
        else:
            return NotImplemented

    def __repr__(self):
        return f"MockDataFrame({self.rows})"

    def to_dataframe(self) -> DataFrame:
        """Transforms this MockDataFrame into a regular pyspark.sql.DataFrame

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> MockDataFrame([MockRow(a=1, b=[2], c={3})]).to_dataframe().show()
        +---+---+---+
        |  a|  b|  c|
        +---+---+---+
        |  1|[2]|[3]|
        +---+---+---+
        <BLANKLINE>

        >>> MockDataFrame({MockRow(a=1, b=[2], c={3})}).to_dataframe().show()
        +---+---+---+
        |  a|  b|  c|
        +---+---+---+
        |  1|[2]|[3]|
        +---+---+---+
        <BLANKLINE>

        :return:
        """
        spark = get_spark_session()
        return spark.createDataFrame([r.to_row() for r in self.rows])



from typing import List

from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from pyspark.sql.types import IntegerType
from pyspark.sql.window import Window

from karadoc.common import spark_utils
from karadoc.common.spark_utils import quote
from karadoc.common.utils.assert_utils import assert_true


def _date_cols_to_numeric(df: DataFrame, reference_time_col: str, ignore_columns: List[str]) -> DataFrame:
    """Replace all date columns with the time difference in days between them and the reference time.

    :param df: a DataFrame
    :param reference_time_col: name of the column to use as reference time
    :param ignore_columns: List of column names to ignore
    :return:
    """
    df_res = df
    date_types = ["timestamp", "date"]
    date_cols = [
        col_name
        for col_name, type in df.dtypes
        if type in date_types and col_name not in [reference_time_col] + ignore_columns
    ]
    for col in date_cols:
        df_res = df_res.withColumn(col, f.datediff(f.col(reference_time_col), f.col(col)))

    return df_res


def _get_always_null_columns(df: DataFrame) -> List[str]:
    """Compute the list of all columns that are always null.

    Example:
    >>> spark = spark_utils.get_spark_session("doctest")
    >>> df = spark.createDataFrame([
    ...         (1, 'a', None, None),
    ...         (2, 'b', None, None),
    ...         (3, 'c', None, None)
    ...     ], "id INT, c1 STRING, c2 STRING, c3 STRING"
    ... )
    >>> always_null_cols = _get_always_null_columns(df)
    >>> always_null_cols.sort()
    >>> print(always_null_cols)
    ['c2', 'c3']

    :param df: a DataFrame
    :return: a list with the name of the columns that are always null
    """
    count_null_df = df.select([f.sum(f.col(col).isNotNull().astype(IntegerType())).alias(col) for col in df.columns])
    row = count_null_df.collect()[0]
    return [col_name for col_name, count in row.asDict().items() if count == 0]


def _quantile_discretize_numeric_col(df: DataFrame, nb_buckets: int, ignore_columns: List[str]) -> DataFrame:
    """Replace all numerical columns with their value discretized by quantile.
    Every column with a numeric type will be converted to bucketized string column.
    :param df: a DataFrame
    :param nb_buckets: the number of buckets to make
    :return: a DataFrame with the same column names, with numerical columns converted into string buckets.
    """
    from pyspark.ml.feature import QuantileDiscretizer

    num_types = ["bigint", "double", "int", "float"]  # excluding 'boolean'
    num_cols = [col_name for col_name, type in df.dtypes if type in num_types and col_name not in ignore_columns]

    for col, type in df.dtypes:
        if type == "boolean" or col in ignore_columns:
            df = df.withColumn(col, f.col(col).cast("string"))

    df_res = df
    for col in num_cols:
        window = Window.partitionBy(f.col("disc_" + col))
        max_digits = len(str(df_res.agg(f.max(f.abs(df_res[col]))).collect()[0][0]))
        df_res = (
            QuantileDiscretizer(numBuckets=nb_buckets, inputCol=col, outputCol="disc_" + col)
            .fit(df_res)
            .transform(df_res)
            .withColumn("max" + col, f.lpad(f.max(col).over(window), max_digits, "0"))
            .withColumn("min" + col, f.lpad(f.min(col).over(window), max_digits, "0"))
            .withColumn(col, f.concat(f.col("min" + col), f.lit("_to_"), f.col("max" + col)))
            .drop(
                "disc_" + col,
                "max" + col,
                "min" + col,
            )
        )
    return df_res


def _compute_frequencies(df: DataFrame, nb_top_values: int) -> DataFrame:
    """Compute the frequency of all key and values, grouped by cohort.
    Only the most frequent values will be kept, up to `nb_top_values`.
    Other values will be replaced by the null and the flag is_other will be set to true.

    Example:

    >>> spark = spark_utils.get_spark_session("doctest")
    >>> df = spark.createDataFrame([
    ...          ('2017-07-24 00:00:00', 'string_col', 'DOGS'),
    ...          ('2017-07-24 00:00:00', 'string_col', 'BIRDS'),
    ...          ('2017-07-24 00:00:00', 'string_col', None),
    ...          ('2017-07-24 00:00:00', 'always_null_col', None),
    ...          ('2017-07-24 00:00:00', 'always_null_col', None),
    ...          ('2017-07-31 00:00:00', 'string_col', 'CATS'),
    ...          ('2017-07-31 00:00:00', 'string_col', 'CATS'),
    ...          ('2017-07-31 00:00:00', 'string_col', 'CATS'),
    ...     ], "cohort STRING, key STRING, val STRING"
    ... )
    >>> df.orderBy('cohort', 'key', 'val').show()
    +-------------------+---------------+-----+
    |             cohort|            key|  val|
    +-------------------+---------------+-----+
    |2017-07-24 00:00:00|always_null_col| null|
    |2017-07-24 00:00:00|always_null_col| null|
    |2017-07-24 00:00:00|     string_col| null|
    |2017-07-24 00:00:00|     string_col|BIRDS|
    |2017-07-24 00:00:00|     string_col| DOGS|
    |2017-07-31 00:00:00|     string_col| CATS|
    |2017-07-31 00:00:00|     string_col| CATS|
    |2017-07-31 00:00:00|     string_col| CATS|
    +-------------------+---------------+-----+
    <BLANKLINE>
    >>> _compute_frequencies(df, 1).orderBy('cohort', 'key', 'val', 'is_other').show()
    +-------------------+---------------+----+--------+---------+------------------+
    |             cohort|            key| val|is_other|val_count|     val_frequency|
    +-------------------+---------------+----+--------+---------+------------------+
    |2017-07-24 00:00:00|always_null_col|null|   false|        2|               1.0|
    |2017-07-24 00:00:00|     string_col|null|   false|        1|0.3333333333333333|
    |2017-07-24 00:00:00|     string_col|null|    true|        2|0.6666666666666666|
    |2017-07-24 00:00:00|     string_col|CATS|   false|        0|               0.0|
    |2017-07-31 00:00:00|always_null_col|null|   false|        0|              null|
    |2017-07-31 00:00:00|     string_col|null|   false|        0|               0.0|
    |2017-07-31 00:00:00|     string_col|null|    true|        0|               0.0|
    |2017-07-31 00:00:00|     string_col|CATS|   false|        3|               1.0|
    +-------------------+---------------+----+--------+---------+------------------+
    <BLANKLINE>

    :param df: a DataFrame with the following schema: "cohort STRING, key STRING, val STRING"
    :param nb_top_values: The number of most frequent values to keep.
    :return: a DataFrame with the following schema:
        "cohort STRING, key STRING, val STRING, is_other BOOLEAN, val_count BIGINT, val_frequency DOUBLE"
    """
    assert_true(df.columns == ["cohort", "key", "val"], df.columns)
    # nb_top_values = 1
    # df.orderBy('cohort', 'key', 'val').show()
    # +-------------------+---------------+-----+
    # |             cohort|            key|  val|
    # +-------------------+---------------+-----+
    # |2017-07-24 00:00:00|always_null_col| null|
    # |2017-07-24 00:00:00|always_null_col| null|
    # |2017-07-24 00:00:00|     string_col| null|
    # |2017-07-24 00:00:00|     string_col|BIRDS|
    # |2017-07-24 00:00:00|     string_col| DOGS|
    # |2017-07-31 00:00:00|     string_col| CATS|
    # |2017-07-31 00:00:00|     string_col| CATS|
    # |2017-07-31 00:00:00|     string_col| CATS|
    # +-------------------+---------------+-----+

    cohort_key_val_count_df = df.groupBy("cohort", "key", "val").agg(f.count(f.lit(1)).alias("count"))
    assert_true(cohort_key_val_count_df.columns == ["cohort", "key", "val", "count"], cohort_key_val_count_df.columns)
    # cohort_key_val_count_df.orderBy('cohort', 'key', 'val').show()
    # +-------------------+---------------+-----+-----+
    # |             cohort|            key|  val|count|
    # +-------------------+---------------+-----+-----+
    # |2017-07-24 00:00:00|always_null_col| null|    2|
    # |2017-07-24 00:00:00|     string_col| null|    1|
    # |2017-07-24 00:00:00|     string_col|BIRDS|    1|
    # |2017-07-24 00:00:00|     string_col| DOGS|    1|
    # |2017-07-31 00:00:00|     string_col| CATS|    3|
    # +-------------------+---------------+-----+-----+

    most_frequent_values = _get_most_frequent_values(cohort_key_val_count_df, nb_top_values)
    assert_true(most_frequent_values.columns == ["key", "val"], most_frequent_values.columns)
    # most_frequent_values.orderBy('key', 'val').show()
    # +---------------+----+
    # |            key| val|
    # +---------------+----+
    # |always_null_col|null|
    # |     string_col|null|
    # |     string_col|CATS|
    # +---------------+----+

    # We flag all the values that are not part of the most frequent values as 'other'
    join_df = (
        cohort_key_val_count_df.alias("T1")
        .join(most_frequent_values.alias("T2"), f.expr("T1.key = T2.key AND T1.val = T2.val"), "left_outer")
        .selectExpr("T1.cohort", "T1.key", "T2.val", "T2.val IS NULL AND T1.val IS NOT NULL as is_other", "count")
    )

    assert_true(join_df.columns == ["cohort", "key", "val", "is_other", "count"], join_df.columns)
    # join_df.orderBy('cohort', 'key', 'val', 'is_other').show()
    # +-------------------+---------------+----+--------+-----+
    # |             cohort|            key| val|is_other|count|
    # +-------------------+---------------+----+--------+-----+
    # |2017-07-24 00:00:00|always_null_col|null|   false|    2|
    # |2017-07-24 00:00:00|     string_col|null|   false|    1|
    # |2017-07-24 00:00:00|     string_col|null|    true|    1|
    # |2017-07-24 00:00:00|     string_col|null|    true|    1|
    # |2017-07-31 00:00:00|     string_col|CATS|   false|    3|
    # +-------------------+---------------+----+--------+-----+

    # We sum together all the occurrences of others
    sum_df = join_df.groupby(["cohort", "key", "val", "is_other"]).agg(f.sum("count").alias("count"))

    assert_true(sum_df.columns == ["cohort", "key", "val", "is_other", "count"], sum_df.columns)
    # sum_df.orderBy('cohort', 'key', 'val', 'is_other').show()
    # +-------------------+---------------+----+--------+-----+
    # |             cohort|            key| val|is_other|count|
    # +-------------------+---------------+----+--------+-----+
    # |2017-07-24 00:00:00|always_null_col|null|   false|    2|
    # |2017-07-24 00:00:00|     string_col|null|   false|    1|
    # |2017-07-24 00:00:00|     string_col|null|    true|    2|
    # |2017-07-31 00:00:00|     string_col|CATS|   false|    3|
    # +-------------------+---------------+----+--------+-----+

    distinct_key_val_df = sum_df.select("key", "val", "is_other").distinct()

    assert_true(distinct_key_val_df.columns == ["key", "val", "is_other"], distinct_key_val_df.columns)
    # distinct_key_val_df.orderBy('key', 'val', 'is_other').show()
    # +---------------+----+--------+
    # |            key| val|is_other|
    # +---------------+----+--------+
    # |always_null_col|null|   false|
    # |     string_col|null|   false|
    # |     string_col|null|    true|
    # |     string_col|CATS|   false|
    # +---------------+----+--------+

    count_per_cohort_df = (
        cohort_key_val_count_df.groupBy("cohort")
        .agg(f.sum(f.col("count")).alias("cohort_count"))
        .select("cohort", "cohort_count")
    )
    assert_true(count_per_cohort_df.columns == ["cohort", "cohort_count"], count_per_cohort_df.columns)
    # count_per_cohort_df.orderBy('cohort').show()
    # +-------------------+------------+
    # |             cohort|cohort_count|
    # +-------------------+------------+
    # |2017-07-24 00:00:00|           5|
    # |2017-07-31 00:00:00|           3|
    # +-------------------+------------+

    # Make a cross-join to build all possible couples 'cohort', 'key', 'val', 'is_other' to fill the holes
    canvas_df = distinct_key_val_df.crossJoin(count_per_cohort_df).select("cohort", "key", "val", "is_other")

    assert_true(canvas_df.columns == ["cohort", "key", "val", "is_other"], canvas_df.columns)
    # canvas_df.orderBy('cohort', 'key', 'val', 'is_other').show()
    # +-------------------+---------------+----+--------+
    # |             cohort|            key| val|is_other|
    # +-------------------+---------------+----+--------+
    # |2017-07-24 00:00:00|always_null_col|null|   false|
    # |2017-07-24 00:00:00|     string_col|null|   false|
    # |2017-07-24 00:00:00|     string_col|null|    true|
    # |2017-07-24 00:00:00|     string_col|CATS|   false|
    # |2017-07-31 00:00:00|always_null_col|null|   false|
    # |2017-07-31 00:00:00|     string_col|null|   false|
    # |2017-07-31 00:00:00|     string_col|null|    true|
    # |2017-07-31 00:00:00|     string_col|CATS|   false|
    # +-------------------+---------------+----+--------+

    # We perform a full join of the DataFrame that counts the occurrence of each values
    # with the canvas DataFrame to fill the holes.
    # We also flag the non-null values that are not in the canvas as not being part of the most frequent values
    complete_df = (
        sum_df.alias("T1")
        .join(
            canvas_df.alias("T2"),
            f.expr(
                "T1.cohort <=> T2.cohort AND T1.key <=> T2.key AND T1.val <=> T2.val AND T1.is_other <=> T2.is_other"
            ),
            "right_outer",
        )
        .selectExpr(
            "COALESCE(T1.cohort, T2.cohort) as cohort",
            "COALESCE(T1.key, T2.key) as key",
            "COALESCE(T1.val, T2.val) as val",
            "COALESCE(T1.is_other, T2.is_other) as is_other",
            "COALESCE(T1.count, 0) as val_count",
        )
    )

    assert_true(complete_df.columns == ["cohort", "key", "val", "is_other", "val_count"], complete_df.columns)
    # complete_df.orderBy('cohort', 'key', 'val', 'is_other').show()
    # +-------------------+---------------+----+--------+---------+
    # |             cohort|            key| val|is_other|val_count|
    # +-------------------+---------------+----+--------+---------+
    # |2017-07-24 00:00:00|always_null_col|null|   false|        2|
    # |2017-07-24 00:00:00|     string_col|null|   false|        1|
    # |2017-07-24 00:00:00|     string_col|null|    true|        2|
    # |2017-07-24 00:00:00|     string_col|CATS|   false|        0|
    # |2017-07-31 00:00:00|always_null_col|null|   false|        0|
    # |2017-07-31 00:00:00|     string_col|null|   false|        0|
    # |2017-07-31 00:00:00|     string_col|null|    true|        0|
    # |2017-07-31 00:00:00|     string_col|CATS|   false|        3|
    # +-------------------+---------------+----+--------+---------+

    frequency_df = complete_df.select(
        "cohort",
        "key",
        "val",
        "is_other",
        "val_count",
        (f.col("val_count") / f.sum("val_count").over(Window.partitionBy("cohort", "key"))).alias("val_frequency"),
    )
    assert_true(
        frequency_df.columns
        == [
            "cohort",
            "key",
            "val",
            "is_other",
            "val_count",
            "val_frequency",
        ],
        frequency_df.columns,
    )
    # frequency_df.orderBy('cohort', 'key', 'val', 'is_other').show()
    # +-------------------+---------------+----+--------+---------+------------------+
    # |             cohort|            key| val|is_other|val_count|     val_frequency|
    # +-------------------+---------------+----+--------+---------+------------------+
    # |2017-07-24 00:00:00|always_null_col|null|   false|        2|               1.0|
    # |2017-07-24 00:00:00|     string_col|null|   false|        1|0.3333333333333333|
    # |2017-07-24 00:00:00|     string_col|null|    true|        2|0.6666666666666666|
    # |2017-07-24 00:00:00|     string_col|CATS|   false|        0|               0.0|
    # |2017-07-31 00:00:00|always_null_col|null|   false|        0|              null|
    # |2017-07-31 00:00:00|     string_col|null|   false|        0|               0.0|
    # |2017-07-31 00:00:00|     string_col|null|    true|        0|               0.0|
    # |2017-07-31 00:00:00|     string_col|CATS|   false|        3|               1.0|
    # +-------------------+---------------+----+--------+---------+------------------+
    # The null here is okay, as this is not supposed to happen in practice with a well-formed DataFrame
    # It happens because there is now value for key 'always_null_col' in cohort '2017-07-31 00:00:00',
    # but in practice, if the cohort exists then there is at least one value for this key.

    return frequency_df


def _get_most_frequent_values(df: DataFrame, nb_top_values: int) -> DataFrame:
    """Compute the most frequent values for each column

    Example:

    >>> spark = spark_utils.get_spark_session("doctest")
    >>> df = spark.createDataFrame([
    ...          ('2017-07-24 00:00:00', 'always_null_col', None, 2),
    ...          ('2017-07-24 00:00:00', 'string_col', 'DOGS', 2),
    ...          ('2017-07-24 00:00:00', 'string_col', None, 1),
    ...          ('2017-07-31 00:00:00', 'string_col', 'CATS', 3),
    ...     ], "cohort STRING, key STRING, val STRING, count INT"
    ... )
    >>> df.show()
    +-------------------+---------------+----+-----+
    |             cohort|            key| val|count|
    +-------------------+---------------+----+-----+
    |2017-07-24 00:00:00|always_null_col|null|    2|
    |2017-07-24 00:00:00|     string_col|DOGS|    2|
    |2017-07-24 00:00:00|     string_col|null|    1|
    |2017-07-31 00:00:00|     string_col|CATS|    3|
    +-------------------+---------------+----+-----+
    <BLANKLINE>
    >>> _get_most_frequent_values(df, 1).orderBy('key', 'val').show()
    +---------------+----+
    |            key| val|
    +---------------+----+
    |always_null_col|null|
    |     string_col|null|
    |     string_col|CATS|
    +---------------+----+
    <BLANKLINE>

    :param df: a DataFrame with the following schema: "cohort STRING, key STRING, val STRING, count INT"
    :param nb_top_values: the max number of values to keep. Nulls values are kept.
    :return: a new DataFrame
    """
    # Example:
    # nb_top_values = 1
    assert_true(df.columns == ["cohort", "key", "val", "count"])
    key_val_count_df = df.groupBy("key", "val").agg(f.sum("count").alias("count"))
    # key_val_count_df.show()
    # +---------------+----+-----+
    # |            key| val|count|
    # +---------------+----+-----+
    # |     string_col|null|    1|
    # |     string_col|DOGS|    2|
    # |     string_col|CATS|    3|
    # |always_null_col|null|    2|
    # +---------------+----+-----+

    # null values should not count in the top most frequent value ranking
    non_null_val_count_col = f.when(f.col("val").isNull(), f.lit(None)).otherwise(f.col("count"))

    window = Window.partitionBy("key").orderBy(non_null_val_count_col.desc())
    most_frequent_values_df = key_val_count_df.withColumn("row_number", f.row_number().over(window)).where(
        (f.col("row_number") <= f.lit(nb_top_values)) | (f.col("val").isNull())
    )

    # most_frequent_values_df.show()
    # +---------------+----+-----+----------+
    # |            key| val|count|row_number|
    # +---------------+----+-----+----------+
    # |     string_col|CATS|    3|         1|
    # |     string_col|null|    1|         3|
    # |always_null_col|null|    2|         1|
    # +---------------+----+-----+----------+

    return most_frequent_values_df.drop("count").drop("row_number")


def _prefix_columns(df: DataFrame) -> DataFrame:
    """Add an underscore prefix to all columns of the given DataFrame
    This is useful to avoid naming collisions

    Example:
    >>> spark = spark_utils.get_spark_session("doctest")
    >>> df = spark.createDataFrame([], "a STRING, b STRING, c STRING, d INT")
    >>> _prefix_columns(df).columns
    ['_a', '_b', '_c', '_d']

    :param df: a DataFrame
    :return: another DataFrame with column renamed
    """
    res = df
    for col in df.columns:
        res = res.withColumnRenamed(col, "_" + col)
    return res


def _prepare_dataframe(df: DataFrame, cohort_cols: List[str]) -> DataFrame:
    """Transform the given Dataframe to make it usable in the analyzis algorithm.
    All the column names are first prefixed with an underscore.
    Then all the cohort columns are wrapped together in a struct column called 'cohort'.

    :param df: a DataFrame
    :param cohort_cols: names of the columns that will be used as cohorts
    :return:
    """
    df2 = _prefix_columns(df)
    cohort_cols = ["_" + col for col in cohort_cols]
    df3 = df2.select(
        f.struct(*[f.col(quote(col)).alias(col[1:]) for col in cohort_cols]).alias("cohort"),
        *[col for col in df2.columns if col not in cohort_cols],
    )
    return df3


def _finalize_dataframe(df: DataFrame) -> DataFrame:
    """Unwraps the cohort column.

    :param df: a DataFrame
    :return:
    """
    return df.select("cohort.*", *df.columns[1:])


def analyse_timeline(df: DataFrame, reference_time_col: str, cohort_cols: List[str], nb_buckets: int = 5) -> DataFrame:
    """Perform an analysis on the given DataFrame that computes the distribution with respect
        to the columns in the 'cohort_cols' list of each other column of the DataFrame.

        Each non-cohort column will be discretized with the following algorithm:
        - Date and Datetime columns are converted in numbers as the difference in days with the `reference_time_col`.
        - Numerical columns are discretized using a Quantile Discretizer algorithm that generates `nb_buckets` buckets.
          They are then cast into strings representing the bucket ranges they fall into.
        - For discretized columns like string, only the top `nb_buckets` frequent values (and the null values) are
          kept as buckets. All remaining values are aggregated a null with the column is_other set to true.

        We then return a DataFrame that gives for each column name, each cohort, each value and each is_other (boolean)
        their number of occurrences and relative frequency relatively to other values in that cohort.
        When a value doesn't appear in a cohort group, a row is still added with a 0 count, to allow for easy plotting.

    Example:
    >>> spark = spark_utils.get_spark_session("doctest")
    >>> df = spark.createDataFrame([
    ...          ('2017-07-24 00:00:00', 'DOGS',  1, None),
    ...          ('2017-07-24 00:00:00', 'DOGS',  1, None),
    ...          ('2017-07-24 00:00:00', 'CATS',  2, None),
    ...          ('2017-07-24 00:00:00', 'BIRDS', 2, None),
    ...          ('2017-07-24 00:00:00', None,    3, None),
    ...          ('2017-07-31 00:00:00', 'CATS',  4, None),
    ...          ('2017-07-31 00:00:00', 'CATS',  5, None),
    ...          ('2017-07-31 00:00:00', 'CATS',  6, None),
    ...     ], "cohort STRING, col_string STRING, col_int INT, col_null STRING"
    ... )
    >>> df.show()
    +-------------------+----------+-------+--------+
    |             cohort|col_string|col_int|col_null|
    +-------------------+----------+-------+--------+
    |2017-07-24 00:00:00|      DOGS|      1|    null|
    |2017-07-24 00:00:00|      DOGS|      1|    null|
    |2017-07-24 00:00:00|      CATS|      2|    null|
    |2017-07-24 00:00:00|     BIRDS|      2|    null|
    |2017-07-24 00:00:00|      null|      3|    null|
    |2017-07-31 00:00:00|      CATS|      4|    null|
    |2017-07-31 00:00:00|      CATS|      5|    null|
    |2017-07-31 00:00:00|      CATS|      6|    null|
    +-------------------+----------+-------+--------+
    <BLANKLINE>
    >>> analyse_timeline(df, 'cohort', ['cohort'], 2).orderBy('cohort', 'key', 'val', 'is_other').show()
    +-------------------+----------+------+--------+---------+-------------+
    |             cohort|       key|   val|is_other|val_count|val_frequency|
    +-------------------+----------+------+--------+---------+-------------+
    |2017-07-24 00:00:00|   col_int|1_to_1|   false|        2|          0.4|
    |2017-07-24 00:00:00|   col_int|2_to_6|   false|        3|          0.6|
    |2017-07-24 00:00:00|  col_null|  null|   false|        5|          1.0|
    |2017-07-24 00:00:00|col_string|  null|   false|        1|          0.2|
    |2017-07-24 00:00:00|col_string|  null|    true|        1|          0.2|
    |2017-07-24 00:00:00|col_string|  CATS|   false|        1|          0.2|
    |2017-07-24 00:00:00|col_string|  DOGS|   false|        2|          0.4|
    |2017-07-31 00:00:00|   col_int|1_to_1|   false|        0|          0.0|
    |2017-07-31 00:00:00|   col_int|2_to_6|   false|        3|          1.0|
    |2017-07-31 00:00:00|  col_null|  null|   false|        3|          1.0|
    |2017-07-31 00:00:00|col_string|  null|   false|        0|          0.0|
    |2017-07-31 00:00:00|col_string|  null|    true|        0|          0.0|
    |2017-07-31 00:00:00|col_string|  CATS|   false|        3|          1.0|
    |2017-07-31 00:00:00|col_string|  DOGS|   false|        0|          0.0|
    +-------------------+----------+------+--------+---------+-------------+
    <BLANKLINE>

    :param df: a DataFrame
    :param reference_time_col: The name of the column to use to normalize other date and datetime columns
    :param cohort_cols: The name of the columns to use as groups for the analysis
    :param nb_buckets: The number of distinct values to keep per column (nulls and 'OTHER' will also appear as values)
    :return: a DataFrame with the following columns: cohort_cols + ['key', 'val', 'val_count', 'val_frequency']
    """

    assert_true(
        reference_time_col in df.columns,
        "%s was not found in the dataframe's column: %s"
        % (
            reference_time_col,
            df.columns,
        ),
    )
    for col in cohort_cols:
        assert_true(col in df.columns, "%s was not found in the dataframe's column: %s" % (col, df.columns))

    df_res1 = _prepare_dataframe(df, cohort_cols)
    reference_time_col = "_" + reference_time_col

    df_res2 = _date_cols_to_numeric(df_res1, reference_time_col, cohort_cols)
    always_null_columns = _get_always_null_columns(df_res2)
    df_res3 = _quantile_discretize_numeric_col(df_res2, nb_buckets, always_null_columns).drop(reference_time_col)

    # Once the unpivot is done, we remove the underscore that we appended to the column names
    df_res4 = spark_utils.unpivot(df_res3, ["cohort"], key_alias="key", value_alias="val").withColumn(
        "key", f.expr("SUBSTR(key, 2)")
    )
    df_res5 = _compute_frequencies(df_res4, nb_buckets)

    df_res6 = _finalize_dataframe(df_res5)

    assert_true(df_res6.columns == cohort_cols + ["key", "val", "is_other", "val_count", "val_frequency"])
    return df_res6



import dynaconf

AIRFLOW_GROUP = "airflow"


def get_airflow_dag_folder_location() -> str:
    return dynaconf.settings.get(f"{AIRFLOW_GROUP}.dag_folder", default="airflow/dags")


def get_airflow_plugin_folder_location() -> str:
    return dynaconf.settings.get(f"{AIRFLOW_GROUP}.plugin_folder", default="airflow/plugins")



from enum import Enum
from typing import Callable, Optional

from pyspark.sql import DataFrame


class CheckSeverity(str, Enum):
    Critical = "Critical"
    Error = "Error"
    Warning = "Warning"
    Info = "Info"
    Debug = "Debug"


class Check:
    global_check_counter = 0

    def __init__(self, func: Callable, description: str, severity: str, name: Optional[str] = None):
        self.func = func
        if name is None:
            self.name = func.__name__
        else:
            self.name = name
        self.description = description
        self.severity = severity

        Check.global_check_counter += 1
        self._creation_rank = Check.global_check_counter
        """Incremental id used to sort the checks in the order they were declared in the QUALITY_CHECK.py file.

        WARNING: When looking at check results of multiple evaluation, the only guarantee is that if you sort
        all the checks with the same table and evaluation_date, they will be sorted in the same order as in the
        QUALITY_CHECK.py file.
        However, there is no guarantee that the id of a given check will remain the same across time.
        """

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class Alert(Check):
    pass


class Metric(Check):
    pass


def alert(description: str, severity: str, name=None) -> Callable[[Callable[[], DataFrame]], Alert]:
    def decorator(func: Callable[[], DataFrame]) -> Alert:
        return Alert(func, description, severity, name)

    return decorator


def metric(description: str, severity: str, name=None) -> Callable[[Callable[[], DataFrame]], Metric]:
    def decorator(func: Callable[[], DataFrame]) -> Metric:
        return Metric(func, description, severity, name)

    return decorator



from typing import Optional

from azure.common.credentials import ServicePrincipalCredentials
from azure.keyvault import KeyVaultAuthentication, KeyVaultClient

from karadoc.common import conf


def get_secret(  # nosec B107
    conn_name: str, secret_id: str, secret_version: str = "", env: Optional[str] = None
) -> str:
    """
    Retrieve a secret from the specified keyvault environment

    Example of keyvault secret store conf to put in the file ".secrets.toml" (should not be tracked in git) :

    >>>
    [development.secret.my_keyvault_env]
      type = "keyvault"
      client_id = "ffffffff-ffff-ffff-ffff-ffffffffffff"
      client_secret = "XXXXXXXXXXXXXXXXXXXXXXX"
      tenant_id = "dddddddd-dddd-dddd-dddd-dddddddddddd"
      url = "https://mykeyvault.vault.azure.net/"

    Example of connection conf to put in file "settings.toml" that will retrieve a secret from keyvault :

    >>>
    [development.connection.my_env]
        login = "some_login"
        password.secret.my_keyvault_env = "secret_id"


    :param conn_name: Name of the keyvault secret store
    :param secret_id: Id of the keyvault secret to retrieve
    :param secret_version: (Optional) version of the secret to retrieve
    :param env: (Optional) Execution environment
    :return: the secret value
    """
    keyvault_conf = conf.get_vault_conf(conn_name, env)

    def auth_callback(server, resource, scope):
        credentials = ServicePrincipalCredentials(
            client_id=keyvault_conf["client_id"],
            secret=keyvault_conf["client_secret"],
            tenant=keyvault_conf["tenant_id"],
            resource="https://vault.azure.net",
        )
        token = credentials.token
        return token["token_type"], token["access_token"]

    client = KeyVaultClient(KeyVaultAuthentication(auth_callback))

    secret_bundle = client.get_secret(keyvault_conf["url"], secret_id, secret_version)
    return secret_bundle.value



from typing import Dict, Optional

from karadoc.common import conf
from karadoc.common.job_core.has_spark import HasSpark


def init_job(job: HasSpark, raw_args: str, spark_conf: Optional[Dict] = None):
    job.init(app_name=f"{conf.APPLICATION_NAME.lower()} {raw_args}", spark_conf=spark_conf)



from enum import IntEnum
from typing import Any, Iterable, List

from termcolor import colored


class ValidationSeverity(IntEnum):
    Critical = 5
    Error = 4
    Warning = 3
    Info = 2
    Debug = 1

    @property
    def repr_color(self) -> str:
        """Color used to represent this severity"""
        return ValidationSeverityColorMap[self]

    @property
    def repr_item(self) -> str:
        """Item name used to represent this severity"""
        return ValidationSeverityStringMap[self]

    def decrease(self) -> "ValidationSeverity":
        """Return the severity one level below this one"""
        return ValidationSeverity(max(self.value - 1, 1))

    def increase(self) -> "ValidationSeverity":
        """Return the severity one level above this one"""
        return ValidationSeverity(min(self.value + 1, 5))


ValidationSeverityColorMap = {
    ValidationSeverity.Critical: "red",
    ValidationSeverity.Error: "red",
    ValidationSeverity.Warning: "yellow",
    ValidationSeverity.Info: "white",
    ValidationSeverity.Debug: "white",
}

ValidationSeverityStringMap = {
    ValidationSeverity.Critical: "critical error",
    ValidationSeverity.Error: "error",
    ValidationSeverity.Warning: "warning",
    ValidationSeverity.Info: "info item",
    ValidationSeverity.Debug: "debug item",
}


class ValidationResult:
    def __init__(self, check_type: str, message: str, default_severity: ValidationSeverity):
        self.check_type = check_type
        self.message = message
        self.severity = default_severity

    def __str__(self) -> str:
        return f"{self.severity.name}: {self.check_type}: {self.message}"

    def __repr__(self) -> str:
        return f'ValidationResult("{self.check_type}", "{self.message}", ValidationSeverity.{self.severity.name})'

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ValidationResult):
            return (self.check_type, self.message, self.severity) == (other.check_type, other.message, other.severity)
        else:
            return NotImplemented

    def color_str(self) -> str:
        return colored(self.__str__(), self.severity.repr_color)


class ValidationResultTemplate:
    """Template class for a Validation result.

    Example:
    >>> template = ValidationResultTemplate("karadoc.demo", "this is a {arg}", ValidationSeverity.Info)
    >>> template(arg="demo")
    ValidationResult("karadoc.demo", "this is a demo", ValidationSeverity.Info)

    """

    def __init__(self, check_type: str, message_template: str, default_severity: ValidationSeverity):
        self.check_type = check_type
        self.message_template = message_template
        self.default_severity = default_severity

    def __call__(self, **kwargs: Any) -> ValidationResult:
        return ValidationResult(self.check_type, self.message_template.format(**kwargs), self.default_severity)

    def __str__(self) -> str:
        return f"{self.default_severity.name}: {self.message_template}"

    def __repr__(self) -> str:
        return (
            f"ValidationResultTemplate("
            f'"{self.check_type}", "{self.message_template}", ValidationSeverity.{self.default_severity.name})'
        )


class ValidationException(Exception):
    def __init__(self, results: List[ValidationResult], severity: ValidationSeverity) -> None:
        plural = "s" if len(results) > 1 else ""
        msg = f"Found the following {severity.repr_item}{plural}:\n" + "\n".join([str(r) for r in results])
        Exception.__init__(self, msg)


def fail_if_results(
    results: Iterable[ValidationResult], severity: ValidationSeverity = ValidationSeverity.Error
) -> None:
    """Raise an exception if the given list contains at least one result of the given severity or higher.

    :param results:
    :param severity:
    :return:
    """
    results = [r for r in results if severity is None or r.severity >= severity]
    if len(results) > 0:
        raise ValidationException(results, severity)



import copy
import getpass
import threading
import uuid
from dataclasses import dataclass
from typing import Dict, Optional

from karadoc.common import conf
from karadoc.common.conf import APPLICATION_NAME

COMMAND_EXECUTION_UUID = str(uuid.uuid4())


@dataclass(repr=False)
class LoggingContext:
    """Collection of values that can be set during execution and will be sent in every log event

    Example:
    observability.get_logging_context() returns a thread-local context that can be modified
    >>> from karadoc.common.observability import logging_context, LogEvent
    >>> logging_context.get_logging_context().model_id = "s.t1"

    The set values will appear in the LogEvent
    >>> LogEvent(message="Hello!", event_type="test").context.model_id
    's.t1'
    """

    application_name: Optional[str] = None
    "Typical value: Karadoc"
    environment: Optional[str] = None
    "Typical value: PR"
    user_name: Optional[str] = None
    "Typical value: JohnDoe"
    command_uuid: Optional[str] = None
    "Unique identifier of the current running command"
    job_uuid: Optional[str] = None
    "Unique identifier of the current running job"
    thread_uuid: Optional[str] = None
    "Unique identifier of the current thread"
    command: Optional[str] = None
    "Typical value: 'run --tables s.t1 s.t2'"
    model_id: Optional[str] = None
    "Typical value: 's.t1'"
    job_type: Optional[str] = None
    "Typical value: SparkJobBatch, QualityCheck, ..."
    job_vars: Optional[Dict[str, str]] = None
    "Typical value: {'day':'YYYY-MM-DD'}"

    def __repr__(self):
        key_values_str = ", ".join(
            [
                f"{key}={repr(value)}"
                for key, value in self.__dict__.items()
                if value is not None and key not in ["thread_id"]
            ]
        )
        res = f"{self.__class__.__name__}({key_values_str})"
        return res


__MAIN_LOGGING_CONTEXT = LoggingContext()
__THREAD_LOCAL_LOGGING_CONTEXT = threading.local()


def get_logging_context() -> LoggingContext:
    """Get a thread-local LoggingContext that can be modified and will be automatically be passed
    to every LogEvent created.

    Example:
    observability.get_logging_context() returns a thread-local context that can be modified
    >>> from karadoc.common.observability import logging_context, LogEvent
    >>> logging_context.get_logging_context().model_id = "s.t1"

    The set values will appear in the LogEvent
    >>> LogEvent(message="Hello!", event_type="test").context.model_id
    's.t1'
    """
    if threading.current_thread().ident == threading.main_thread().ident:
        if __MAIN_LOGGING_CONTEXT.command_uuid is None:
            __MAIN_LOGGING_CONTEXT.command_uuid = str(uuid.uuid4())
            __MAIN_LOGGING_CONTEXT.environment = conf.get_env()
            __MAIN_LOGGING_CONTEXT.user_name = getpass.getuser()
            __MAIN_LOGGING_CONTEXT.application_name = APPLICATION_NAME
            __MAIN_LOGGING_CONTEXT.thread_id = str(uuid.uuid4())
        return __MAIN_LOGGING_CONTEXT
    if not hasattr(__THREAD_LOCAL_LOGGING_CONTEXT, "value"):
        __THREAD_LOCAL_LOGGING_CONTEXT.value = copy.deepcopy(__MAIN_LOGGING_CONTEXT)
        __THREAD_LOCAL_LOGGING_CONTEXT.value.thread_id = str(uuid.uuid4())
    return __THREAD_LOCAL_LOGGING_CONTEXT.value



from unittest import mock

from dynaconf import settings
from dynaconf.utils.boxing import DynaBox

from karadoc.common.conf.conf_box import dict_to_dynabox

original_spark_settings = dict_to_dynabox({"spark": {"conf": settings.get("spark.conf")}})


# Explanation
# -----------
# This method is necessary to simulate dynaconf's slightly weird behavior.
#
# For instance, if I have the following settings.toml
#
#       [development]
#       a.b.c = "x"
#
# And if I run the following code:
#
#       from dynaconf import settings
#
#       print(settings.get("a").get("b").get("c") == "x")
#       # True
#       print(settings.get("a.b").get("c") == "x")
#       # True
#       print(settings.get("a").get("b.c") == None)
#       # True
#
# We notice that the keys are split by "." at the root level (settings) but not at the subsequent levels (Dynabox)
def _split_get(box, key: str):
    """Get a key from a nested dict-like object by splitting the key according to "."

    >>> b = {"a": {"b": 2}}
    >>> _split_get(b, "a")
    {'b': 2}
    >>> _split_get(b, "a.b")
    2
    >>> _split_get(b, "c")

    >>> _split_get(b, "a.c")

    >>> _split_get(b, "c.c")

    >>> _split_get({}, "a")

    :param box: A nested-dict-like object
    :param key: A string key
    :return: the value located at entry key
    """
    split_key = key.split(".")
    res = box
    for k in split_key:
        res = res.get(k)
        if res is None:
            return None
    return res


def __merge_2_boxes(box1: DynaBox, box2: DynaBox) -> DynaBox:
    """Merges two DynaBox together

    >>> box1 = dict_to_dynabox({"a": {"b": 2}})
    >>> box2 = dict_to_dynabox({"a": {"b": 3, "c": 3}})
    >>> __merge_2_boxes(box1, box2)
    <Box: {'a': {'b': 2, 'c': 3}}>

    :param box1:
    :param box2:
    :return:
    """
    # We only need the keys, but we use a dict instead of a set because dicts preserve insertion order on keys
    keys = {**box1, **box2}
    res = {}
    for key, _ in keys.items():
        v1 = box1.get(key)
        v2 = box2.get(key)
        if isinstance(v1, DynaBox) and isinstance(v2, DynaBox):
            res[key] = merge_boxes(v1, v2)
        elif key in box1:
            res[key] = v1
        else:
            res[key] = v2
    return dict_to_dynabox(res)


def merge_boxes(box: DynaBox, *boxes: DynaBox) -> DynaBox:
    """Recursively merge multiple DynaBoxes together
    The left-most boxes take precedence.

    >>> box1 = dict_to_dynabox({"a": {"b": 2}})
    >>> box2 = dict_to_dynabox({"a": {"b": 3, "c": 3}})
    >>> box3 = dict_to_dynabox({"d": 4})
    >>> merge_boxes(box1, box2, box3)
    <Box: {'a': {'b': 2, 'c': 3}, 'd': 4}>

    :param box: The first box (we need at least one argument)
    :param boxes: The other boxes
    :return:
    """
    res = box
    for b in boxes:
        res = __merge_2_boxes(res, b)
    return res


def _coalesce_box_get(*boxes: DynaBox):
    """Returns a "get(key, default)" method that returns the value found in the first dictionary that contains
    that key, or default otherwise.
    If the returned value is a DynaBox, it is merged with the values found in the other DynaBoxes.


    This also splits the key on "." and recursively look for nested entries.

    >>> box1 = dict_to_dynabox({"a": {"b": 2}})
    >>> box2 = dict_to_dynabox({"a": {"b": 3, "c": 3}})
    >>> box3 = dict_to_dynabox({"d": 4})
    >>> cget = _coalesce_box_get(box1, box2, box3)
    >>> cget("a.b")
    2

    The first non-null value matching the key is returned
    >>> cget("a.c")
    3

    Except for values of type Dict, which are merged for the whole dicts and returned as a DynaBox
    >>> cget("a")
    <Box: {'b': 2, 'c': 3}>
    >>> cget("d")
    4
    >>> cget("a.missing", 0)
    0

    When merging two values of incompatible type, a TypeError is raised
    >>> box1 = dict_to_dynabox({"a": {"b": 2}})
    >>> box2 = dict_to_dynabox({"a": 2})
    >>> cget = _coalesce_box_get(box1, box2, box3)
    >>> cget("a")
    Traceback (most recent call last):
      ...
    TypeError: key a matches values of incompatible types: {'b': 2} and 2

    :param dicts:
    :return:
    """

    def get(key: str, default=None):
        """Return the value in the first dictionary that contains this key.

        :param key:
        :param default:
        :return:
        """
        res = None
        dict_res = {}
        for d in boxes:
            res = _split_get(d, key)
            if res is not None:
                if isinstance(res, DynaBox):
                    dict_res = merge_boxes(dict_res, res)
                elif dict_res != {}:
                    raise TypeError(f"key {key} matches values of incompatible types: {dict_res} and {res}")
                else:
                    return res
        if dict_res != {}:
            return dict_to_dynabox(dict_res)
        else:
            return default

    return get


def _override_for_test_class(box: DynaBox, key, default):
    return _coalesce_box_get(box, original_spark_settings)(key, default)


def mock_settings_for_test_class(settings_dict=None):
    """Method used to override Karadoc's settings for testing purposes:

    You can use it as mock in unit tests or directly override the settings in a Python Console.

    Example :

        .. code-block:: python

            from unittest import TestCase
            from unittest import mock
            @mock_settings_for_test_class({'model_dir': 'tests/resources/model_graph'})
            class TestSomething(TestCase):
                ...

    :param settings_dict:
    :return:
    """
    if settings_dict is None:
        settings_dict = {}
    settings_box = dict_to_dynabox(settings_dict)

    def override(key, default=None):
        return _override_for_test_class(settings_box, key, default)

    return mock.patch("dynaconf.settings.get", override)


def mock_settings_for_test(settings_dict):
    """Method used to override Karadoc's settings for testing purposes:

    You can use it as mock in unit tests or directly override the settings in a Python Console.

    Example as unit test :

    Example :

        .. code-block:: python

            from unittest import TestCase
            from unittest import mock
            @mock_settings_for_test_class()
            class TestSomething(TestCase):

                @mock_settings_for_test({'allow_missing_vars': True})
                def test_something():
                ...

    :param settings_dict:
    :return:
    """
    settings_box = dict_to_dynabox(settings_dict)

    def override(original_box, key, default):
        return _coalesce_box_get(settings_box, original_box, original_spark_settings)(key, default)

    return mock.patch("karadoc.test_utils.mock_settings._override_for_test_class", override)



from argparse import ArgumentParser, Namespace
from typing import Dict, Optional

from pyspark.sql import DataFrame

from karadoc.common import conf, run
from karadoc.common.commands.command import Command
from karadoc.common.commands.help import LineBreakHelpFormatter
from karadoc.common.commands.options.dry_option import DryOption
from karadoc.common.commands.options.read_from_option import ReadFromOption
from karadoc.common.commands.options.spark_conf_option import SparkConfOption
from karadoc.common.commands.options.tables_option import TablesOption
from karadoc.common.commands.options.vars_option import VarsOption
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.commands.spark import init_job
from karadoc.common.commands.utils import run_job_with_logging
from karadoc.common.exceptions import JobDisabledException
from karadoc.common.model import variables
from karadoc.common.run.spark_batch_job import SparkBatchJob
from karadoc.common.validations import fail_if_results
from karadoc.common.validations.connection_validations import validate_connections
from karadoc.common.validations.job_validations import (
    validate_inputs,
    validate_output_partition,
)


def inspect_df(df: DataFrame):
    """Function used for unit tests. Mock this function to be able to inspect the produced DataFrame
    during the execution of the command."""
    pass


def inspect_job(job: SparkBatchJob):
    """Function used for unit tests. Mock this function to be able to inspect the SparkBatchJob object
    during the execution of the command."""
    pass


def _run_table(job: SparkBatchJob, args) -> Optional[DataFrame]:
    """Computes the Spark DataFrame build by a SparkBatchJob with the specified args.

    :param job:
    :param args:
    :return:
    """
    df: Optional[DataFrame] = None
    if args.remote_env is not None:
        job._configure_remote_input(args.remote_env)
    run.print_job_properties(job)
    fail_if_results(validate_inputs(job))
    fail_if_results(validate_output_partition(job))
    fail_if_results(validate_connections(job))
    if job.disable:
        if not args.dry:
            raise JobDisabledException()
        else:
            print("WARN: The job has been disabled, it should not be launched")
    if not args.dry:
        init_job(job=job, raw_args=args.raw_args, spark_conf=args.spark_conf)
        if args.limit_inputs is not None:
            job._limit_inputs = args.limit_inputs
        if args.limit_external_inputs is not None:
            job._limit_external_inputs = args.limit_external_inputs
        df = job.run()
        if args.limit_output is not None and df is not None:
            df = df.limit(args.limit_output)
    elif not conf.is_dev_env():
        init_job(job=job, raw_args=args.raw_args, spark_conf=args.spark_conf)
    return df


def _run_model(args: Namespace, model_id: str, job_vars: Dict[str, str], **kwargs):
    job: SparkBatchJob = run.load_runnable_populate(model_id, job_vars)
    df = _run_table(job, args)
    inspect_df(df)
    if df is not None:
        run.write_output(job, df, no_write=args.no_write, no_export=args.no_export)
    inspect_job(job)


class RunCommand(Command):
    description = "run POPULATE.py scripts for the specified tables"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.formatter_class = LineBreakHelpFormatter

        TablesOption.add_arguments(parser)
        DryOption.add_arguments(parser)
        ReadFromOption.add_arguments(parser)
        VarsOption.add_arguments(parser)
        SparkConfOption.add_arguments(parser)

        parser.add_argument(
            "--limit-output",
            dest="limit_output",
            metavar="N",
            type=int,
            help="Only take the first N rows of the DataFrame returned by the run() method. Useful for testing.",
        )
        parser.add_argument(
            "--limit-inputs",
            dest="limit_inputs",
            metavar="N",
            type=int,
            help="Only take the first N rows of the DataFrame of any input loaded. Useful for testing.",
        )
        parser.add_argument(
            "--limit-external-inputs",
            dest="limit_external_inputs",
            metavar="N",
            type=int,
            help="Only take the first N rows of the DataFrame of any external input loaded. Useful for testing.",
        )
        parser.add_argument(
            "--no-export",
            dest="no_export",
            default=False,
            action="store_true",
            help="Disable writing on external outputs. Useful for testing",
        )
        parser.add_argument(
            "--no-write",
            dest="no_write",
            default=False,
            action="store_true",
            help="Disable writing any output or external output. Useful for testing",
        )
        parser.add_argument(
            "--validate",
            dest="validate",
            default=False,
            action="store_true",
            help="Activate validation mode. This is equivalent to setting all the following options:"
            "\n  --limit-inputs 0"
            "\n  --limit-output 0"
            "\n  --limit-external-inputs 0"
            "\n  --no-write"
            "\n  --no-export"
            "\nWARNING: Please make sure the jobs being validated have no side-effects.",
        )

    @staticmethod
    def do_command(args: Namespace) -> ReturnCode:
        return_code = ReturnCode.Success
        vars_list = variables.expand_vars(args.vars)
        if args.validate:
            args.limit_output = 0
            args.limit_inputs = 0
            args.limit_external_inputs = 0
            args.no_write = True
            args.no_export = True
        for table in args.tables:
            for job_vars in vars_list:
                return_code &= run_job_with_logging(
                    func=_run_model, job_type=SparkBatchJob, args=args, model_id=table, job_vars=job_vars
                )

        return return_code



from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from typing import Optional

from karadoc.common.commands.return_code import ReturnCode


class Command(ABC):
    """Implement this class to define new karadoc commands.

    Each command class must be implemented in a separate python file.

    A Command implementation must implement the following elements:

    - description: str.
    - add_arguments(parser: ArgumentParser).
    - do_command(args: Namespace) -> Optional[ReturnCode]:
    """

    command_depth: int
    """Depth of the command, set by the command loader.
      (e.g. in "karadoc build package", 'build' has depth 0 and 'package' has depth 1)
    """

    @classmethod
    def _get_command_module(cls) -> str:
        """Get the name of the module containing the current command"""
        return cls.__module__

    @classmethod
    def _get_command_name(cls) -> str:
        """Get the name of the module containing the current command"""
        return cls._get_command_module().split(".")[-1]

    @property
    @abstractmethod
    def description(self) -> str:
        """Description of the command which will be displayed in the help"""
        pass

    @staticmethod
    @abstractmethod
    def add_arguments(parser: ArgumentParser) -> None:
        """Method used to set the options and arguments of the command via parser.add_argument()"""
        pass

    @staticmethod
    @abstractmethod
    def do_command(args: Namespace) -> Optional[ReturnCode]:
        """Entrypoint of the command.
        This method can return nothing if it's only way of failing is raising an Exception.
        Commands that can fail without raising exceptions (e.g. a validation command) should return a ReturnCode.
        """
        pass



from argparse import ArgumentParser, Namespace

import networkx as nx

from karadoc.common.commands.command import Command
from karadoc.common.commands.options.tables_option import TablesOption
from karadoc.common.graph import table_graph
from karadoc.common.model import table_index
from karadoc.common.validations import fail_if_results
from karadoc.common.validations.graph_validations import validate_graph


class ShowGraphCommand(Command):
    description = "display the dependency graph for the specified tables"

    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        TablesOption.add_arguments(parser)
        parser.add_argument(
            "-b",
            "--before",
            dest="before",
            type=int,
            default=0,
            metavar="DIST",
            nargs="?",
            help="Recursively show upstream dependencies of the specified tables up to a distance of DIST",
        )
        parser.add_argument(
            "-a",
            "--after",
            dest="after",
            type=int,
            default=0,
            metavar="DIST",
            nargs="?",
            help="Recursively show downstream dependencies of the specified tables up to a distance of DIST",
        )
        parser.add_argument(
            "--reduce",
            dest="reduce",
            default=False,
            action="store_true",
            help="Perform a transitive reduction on the graph. For instance if we have A -> B -> C and A -> C, "
            "the dependency A -> C will be dropped",
        )
        parser.add_argument(
            "--show-hidden",
            dest="show_hidden",
            default=False,
            action="store_true",
            help="Display all dependencies, even those hidden by setting `job.hide_inputs = True`",
        )
        parser.add_argument(
            "--output-format",
            dest="output_format",
            default="png",
            type=str,
            help="Output format of the graph (e.g. 'png', 'svg')",
        )
        parser.add_argument(
            "--print-topological-sort", dest="print_topological_sort", default=False, action="store_true", help=""
        )

    @staticmethod
    def do_command(args: Namespace) -> None:
        index = table_index.build_table_index()
        graph = table_graph.build_graph(index)

        if args.reduce:
            graph = nx.transitive_reduction(graph)
        validation_results = validate_graph(graph)

        ignored_edges = []
        if not args.show_hidden:
            ignored_edges = [
                (input_name, table_name)
                for table_name, table in index.items()
                if table.populate is not None and table.populate.hide_inputs
                for input_name in table.populate.input_tables
            ]

        graph_filters = table_graph.build_graph_filters(args.tables, args.before, args.after)
        graph = table_graph.get_filtered_subgraph(graph, graph_filters, ignored_edges)

        tables = [filter.node for filter in graph_filters]
        table_graph.render_graph(graph, index, tables, output_format=args.output_format)

        fail_if_results(validation_results)

        if args.print_topological_sort:
            for t in table_graph.get_topological_sort(graph):
                print(t)



from airflow.models import BaseOperator


def is_trigger(task: BaseOperator) -> bool:
    return task.task_type == "TriggerDagRunOperator"


def is_external_sensor(task: BaseOperator) -> bool:
    return task.task_type == "ExternalTaskSensor"



from argparse import ArgumentParser
from typing import List, Tuple

import pyspark.sql.functions as f
from pyspark.sql import DataFrame

from karadoc.common import Job, run
from karadoc.common.commands.command import Command
from karadoc.common.commands.options.read_from_option import ReadFromOption
from karadoc.common.commands.options.tables_option import TablesOption
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.commands.spark import init_job
from karadoc.common.quality import CheckSeverity, alert
from karadoc.common.quality.exec import execute_alert
from karadoc.common.spark_utils import quote_idempotent, unquote


def _get_key_columns(job: Job):
    """Returns the set of all flat keys used as part of primary or secondary keys. It does not contain duplicate names

    >>> from karadoc.common import Job
    >>> job = Job()
    >>> _get_key_columns(job)
    []
    >>> job.primary_key = 'abc'
    >>> _get_key_columns(job)
    ['abc']
    >>> job.primary_key = ("c1", "c2", "c3")
    >>> _get_key_columns(job)
    ['c1', 'c2', 'c3']
    >>> job.secondary_keys = [("c1", "c4"), "c2"]
    >>> _get_key_columns(job)
    ['c1', 'c2', 'c3', 'c4']

    :param job:
    :return:
    """
    key = job.standardized_primary_key
    if key is None:
        key = ()
    # Hack: unlike sets, dict preserves ordering on keys.
    return list({column: None for tup in [key] + job.standardized_secondary_keys for column in tup}.keys())


def _get_keys(job: Job) -> List[Tuple[str, ...]]:
    """Returns the set of all columns used as part of primary or secondary keys

    >>> from karadoc.common import Job
    >>> job = Job()
    >>> _get_keys(job)
    []
    >>> job.primary_key = 'c1'
    >>> _get_keys(job)
    [('c1',)]
    >>> job.primary_key = ('c1', 'c2')
    >>> _get_keys(job)
    [('c1', 'c2')]
    >>> job.secondary_keys = [('c1', 'c3'), 'c4']
    >>> _get_keys(job)
    [('c1', 'c2'), ('c1', 'c3'), ('c4',)]

    :param job:
    :return:
    """
    key = job.standardized_primary_key
    if key is None:
        key = []
    else:
        key = [key]
    # Hack: unlike sets, dict preserves ordering on keys.
    return list({tup: None for tup in key + job.standardized_secondary_keys}.keys())


class NoKeyDefinedException(Exception):
    """
    Typed exception, to be raised if there is no key defined on this table.
    """

    def __init__(self, table: str):
        """
        :param table: keyless table
        """
        super().__init__("ERROR: No key is defined for the job")
        self.table = table


def _check_key_uniqueness(df: DataFrame, key: Tuple[str, ...], table: str, output_alert_table: str):
    @alert(
        description="Check that the following key has no duplicates: %s" % str(key),
        severity=CheckSeverity.Critical,
        name=f"check_key_uniqueness:{key}",
    )
    def check_key_uniqueness() -> DataFrame:
        key_cols_aliased = [f.col(c).alias(unquote(c)) for c in key]
        key_cols = [quote_idempotent(c) for c in key]

        df1 = df.groupby(*list(key_cols_aliased)).count().filter("count > 1").orderBy(f.desc("count"))
        return df1.select(*list(key_cols), *[f.col("count").alias("nb_duplicates")])

    execute_alert(table, check_key_uniqueness, {}, output_alert_table)


def _check_keys_uniqueness(df: DataFrame, keys: List[Tuple[str, ...]], table: str, output_alert_table: str):
    if len(keys) == 0:
        raise NoKeyDefinedException(table=table)
    for key in keys:
        _check_key_uniqueness(df, key, table, output_alert_table)


def _check_non_null_keys(df: DataFrame, key_columns: List[str], table: str, output_alert_table: str):
    @alert(
        description="Checks that the following columns are never null: %s" % str(key_columns),
        severity=CheckSeverity.Critical,
    )
    def check_non_null_keys():
        """Checks if a key has null values in a dataframe.
        If the key is composite, we check that every element is non-null
        Example:
                key = ('user_id', 'channel_id')
                The function checks that neither user_id nor channel_id has null values
        """
        null_keys_array = f.array(*[f.when(f.col(k).isNull(), f.lit(k)) for k in key_columns])
        res_df = (
            df.select(null_keys_array.alias("null_keys"))
            .select(f.expr("FILTER(null_keys, x -> x IS NOT NULL)").alias("null_keys"))
            .where("SIZE(null_keys) > 0")
        )
        return res_df

    execute_alert(table, check_non_null_keys, {}, output_alert_table)


class CheckKeysCommand(Command):
    description = "check that the table's primary and secondary keys are unique and non-nulls"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        TablesOption.add_arguments(parser)
        ReadFromOption.add_arguments(parser)

        parser.add_argument(
            "--output-alert-table",
            dest="output_alert_table",
            default=None,
            required=False,
            type=str,
            metavar="TABLE",
            help="Write the generated check keys DataFrames to the outputs and external_outputs of the "
            "specified table. This table must have a valid POPULATE.py file, with a run() method that "
            "will be ignored.",
        )

    @staticmethod
    def do_command(args) -> ReturnCode:
        for table in args.tables:
            job = run.load_populate(table)
            init_job(job=job, raw_args=args.raw_args)
            if args.remote_env is not None:
                job._configure_remote_input(args.remote_env)
            print("Check keys started for table %s" % table)

            job.inputs = {"source": table}
            df = job.read_table("source")

            keys = _get_keys(job)
            _check_keys_uniqueness(df, keys, table, args.output_alert_table)

            key_columns = _get_key_columns(job)
            _check_non_null_keys(df, key_columns, table, args.output_alert_table)

        return ReturnCode.Success



from io import StringIO
from typing import Dict, Optional

from pyspark.sql import Column, DataFrame, SparkSession
from pyspark.sql import functions as f
from termcolor import colored

from karadoc.common import spark_utils
from karadoc.common.exceptions.utils import display_exception
from karadoc.common.job_core.load import (
    load_non_runnable_action_file,
    load_runnable_action_file,
)
from karadoc.common.quality.checks import Alert, Metric
from karadoc.common.quality.quality_check_job import QualityCheckJob
from karadoc.common.run import exec
from karadoc.common.run.exec import load_runnable_populate


def load_quality_check(full_table_name) -> QualityCheckJob:
    """Return the `job` object declared in the QUALITY_CHECK.py file of the given table with its default variables.

    This method is useful for inspecting a QUALITY_CHECK.py file without running it.
    For running quality checks, please use `load_runnable_quality_check` instead.

    :param full_table_name: Full name of the table (schema_name.table_name)
    :return: the QualityCheckJob object defined in the QUALITY_CHECK.py file
    """
    return load_non_runnable_action_file(full_table_name, QualityCheckJob)


def load_runnable_quality_check(full_table_name, passed_vars: Optional[Dict[str, str]]) -> QualityCheckJob:
    """Return the `job` object declared in the QUALITY_CHECK.py file of the given table, override its variables
    and check the load variables match those declared in the QUALITY_CHECK.py file.

    :param full_table_name: Full name of the table (schema_name.table_name)
    :param passed_vars: Set the variables of the job and check that they match the ones declared
    :return: the QualityCheckJob object defined in the QUALITY_CHECK.py file
    """
    return load_runnable_action_file(full_table_name, QualityCheckJob, passed_vars)


def __create_alert_struct(table: str, alert: Alert, status: str) -> Column:
    return f.struct(
        f.lit(table).alias("table"),
        f.lit(alert.name).alias("name"),
        f.lit(alert.description).alias("description"),
        f.lit(alert.severity).alias("severity"),
        f.lit(f.current_timestamp()).alias("evaluation_date"),
        f.lit(status).alias("status"),
        f.lit(alert._creation_rank).alias("rank"),
    ).alias("alert")


def __alert_ok_dataframe(spark: SparkSession, table: str, alert: Alert) -> DataFrame:
    alert_col = __create_alert_struct(table, alert, status="ok")
    values_col = spark_utils.empty_array("STRUCT<key: STRING, value: STRING>").alias("columns")
    return spark.sql("SELECT 1").select(alert_col, values_col)


def __alert_exception_dataframe(spark: SparkSession, table: str, alert: Alert, e: Exception) -> DataFrame:
    display_exception(e)
    dest = StringIO()
    display_exception(e, dest)
    exception_message = dest.getvalue()
    alert_col = __create_alert_struct(table, alert, status="error")
    values_col = f.array(f.struct(f.lit("exception").alias("key"), f.lit(exception_message).alias("value"))).alias(
        "columns"
    )
    return spark.sql("SELECT 1").select(alert_col, values_col)


def __transform_alert_for_export(table: str, alert: Alert, df: DataFrame) -> DataFrame:
    alert_col = __create_alert_struct(table, alert, status="ko")
    values_col = spark_utils.to_generic_struct(*df.columns, col_name_alias="key", col_value_alias="value").alias(
        "columns"
    )
    return df.select(alert_col, values_col)


def __print_alert(alert: Alert):
    try:
        df = __compute_alert_dataframe(alert)
        df = df.persist()
        count = df.count()
        if count == 0:
            print(colored("%s: OK\n%s" % (alert.name, alert.description), "green"))
        else:
            print(colored("%s: NOT OK (%s rows in error)\n%s" % (alert.name, count, alert.description), "red"))
            df.show(10, False)
    except Exception as e:
        print(colored("Error running alert: %s" % alert.name, "red"))
        display_exception(e)


def __write_alert(table: str, alert: Alert, output_table: str, vars: Dict[str, str]):
    export_job = load_runnable_populate(output_table, vars)
    export_job.init()
    try:
        df = __compute_alert_dataframe(alert)
        df = df.persist()
        df = __transform_alert_for_export(table, alert, df)
        count = df.count()
        if count == 0:
            print(colored("%s: OK\n%s" % (alert.name, alert.description), "green"))
            df = __alert_ok_dataframe(export_job.spark, table, alert)
        else:
            print(colored("%s: NOT OK (%s rows in error)\n%s" % (alert.name, count, alert.description), "red"))
        exec.write_output(export_job, df)
    except Exception as e:
        print(colored("Error running alert: %s" % alert.name, "red"))
        df = __alert_exception_dataframe(export_job.spark, table, alert, e)
        exec.write_output(export_job, df)


def inspect_alert_dataframe(alert: Alert, df: DataFrame):
    """Function used for unit tests. Mock this function to be able to inspect the produced DataFrame
    during the execution of the command."""
    pass


def __compute_alert_dataframe(alert: Alert) -> DataFrame:
    df = alert.func()
    inspect_alert_dataframe(alert, df)
    if not isinstance(df, DataFrame):
        raise TypeError(
            f"Alert {alert.name} is supposed to return a Spark DataFrame, but returned a {type(df)} instead"
        )
    return df


def execute_alert(table: str, alert: Alert, vars: Dict[str, str], output_table: Optional[str]):
    if output_table is not None:
        __write_alert(table, alert, output_table, vars)
    else:
        __print_alert(alert)


def execute_metric(table: str, metric: Metric, vars: Dict[str, str], output_table: Optional[str]):  # NOSONAR
    # TODO: this feature is not fully implemented yet. We should support output_tables.
    print("metric: %s" % metric.description)
    df = metric.func()
    if not isinstance(df, DataFrame):
        raise TypeError(
            f"Metric {metric.name} is supposed to return a Spark DataFrame, but returned a {type(df)} instead"
        )
    df.show(10, False)



import json
import re
from collections import OrderedDict
from copy import copy
from typing import Dict, Generator, Iterable, List, Optional, Set, Tuple, Union, cast

from pyspark.sql import Column, DataFrame, SparkSession
from pyspark.sql import functions as f
from pyspark.sql.types import (
    ArrayType,
    DataType,
    MapType,
    StringType,
    StructField,
    StructType,
)

from karadoc.common.utils.assert_utils import assert_true
from karadoc.spark.utils import get_spark_session  # noqa: F401 (used only in doctest)

OrderedTree = Union["OrderedTree", Dict[str, "OrderedTree"]]


def quote(col: str) -> str:
    """Puts the given column name into quotes.

    This is useful in particular when some column names contains dots,
    which usually happens after using the :func:`flatten` operation.

    Examples:

    >>> quote('a')
    '`a`'
    >>> quote('a.b')
    '`a.b`'

    Columns names can even contain backquotes, by escaping backquotes with another backquote.

    >>> quote('a`b')
    '`a``b`'

    Differs from `quote_idempotent` on this specific case
    >>> quote('`a`')
    '```a```'

    :param col:
    :return:
    """
    return "`" + col.replace("`", "``") + "`"


def quote_idempotent(col: str) -> str:
    """Puts the given column name into quotes if it is not already

    This is useful in particular when some column names contains dots,
    which usually happens after using the :func:`flatten` operation.

    Examples:

    >>> quote_idempotent('a')
    '`a`'
    >>> quote_idempotent('a.b')
    '`a.b`'

    Columns names can even contain backquotes, by escaping backquotes with another backquote.

    >>> quote_idempotent('a`b')
    '`a``b`'

    Differs from `quote` on this specific case
    >>> quote_idempotent('`a`')
    '`a`'

    :param col:
    :return:
    """
    return quote(unquote(col))


def quote_columns(columns: List[str]) -> List[str]:
    """Puts every column name of the given list into quotes.

    This is useful in particular when some column names contains dots,
    which usually happens after using the :func:`flatten` operation.

    :param columns:
    :return:
    """
    return [quote(col) for col in columns]


def unquote(col_name: str) -> str:
    """Removes quotes from a quoted column name.

    Examples:

    >>> unquote('a')
    'a'
    >>> unquote('`a`')
    'a'
    >>> unquote('`a.b`')
    'a.b'

    Columns names can even contain backquotes, by escaping backquotes with another backquote.

    >>> unquote('`a``b`')
    'a`b'

    :param col_name:
    :return:
    """
    if col_name[0] == "`" and col_name[-1] == "`":
        col_name = col_name[1:-1]
    return col_name.replace("``", "`")


def df_schema_as_json_string(df: DataFrame) -> str:
    schema_dict = json.loads(df.schema.json())
    return json.dumps(schema_dict, indent=2, sort_keys=True)


def get_schema_from_json(s: str) -> StructType:
    return StructType.fromJson(json.loads(s))


def extract_json_column(spark: SparkSession, df: DataFrame, columns: Union[str, List[str]]) -> DataFrame:
    """
    Transform the specified columns containing json strings in the given dataframe
    into structs containing the equivalent parsed information.

    This method uses Spark's json reader methods, but is more useful because Spark
    can only read json from an RDD[String]. This method has the advantage of allowing
    to extract several json columns and to keep the other Dataframe columns in the result.

    WARNING : when you use this method on a column that is inside a struct (e.g. column "a.b.c"),
    instead of replacing that column it will create a new column outside the struct (e.g. "`a.b.c`") (See Example 2).

    Example 1 :
    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame([
    ...         (1, '[{"a": 1}, {"a": 2}]'),
    ...         (1, '[{"a": 2}, {"a": 4}]'),
    ...         (2, None)
    ...     ], "id INT, json1 STRING"
    ... )
    >>> df.show()
    +---+--------------------+
    | id|               json1|
    +---+--------------------+
    |  1|[{"a": 1}, {"a": 2}]|
    |  1|[{"a": 2}, {"a": 4}]|
    |  2|                null|
    +---+--------------------+
    <BLANKLINE>
    >>> df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- json1: string (nullable = true)
    <BLANKLINE>
    >>> extract_json_column(spark, df, 'json1').printSchema()
    root
     |-- id: integer (nullable = true)
     |-- json1: array (nullable = true)
     |    |-- element: struct (containsNull = true)
     |    |    |-- a: long (nullable = true)
    <BLANKLINE>

    Example 2 : json inside a struct :
    >>> df = spark.createDataFrame([
    ...         (1, {'json1': '[{"a": 1}, {"a": 2}]'}),
    ...         (1, {'json1': '[{"a": 2}, {"a": 4}]'}),
    ...         (2, None)
    ...     ], "id INT, struct STRUCT<json1: STRING>"
    ... )
    >>> df.show(10, False)
    +---+----------------------+
    |id |struct                |
    +---+----------------------+
    |1  |{[{"a": 1}, {"a": 2}]}|
    |1  |{[{"a": 2}, {"a": 4}]}|
    |2  |null                  |
    +---+----------------------+
    <BLANKLINE>
    >>> df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- struct: struct (nullable = true)
     |    |-- json1: string (nullable = true)
    <BLANKLINE>
    >>> res = extract_json_column(spark, df, 'struct.json1')
    >>> res.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- struct: struct (nullable = true)
     |    |-- json1: string (nullable = true)
     |-- struct.json1: array (nullable = true)
     |    |-- element: struct (containsNull = true)
     |    |    |-- a: long (nullable = true)
    <BLANKLINE>
    >>> res.show(10, False)
    +---+----------------------+------------+
    |id |struct                |struct.json1|
    +---+----------------------+------------+
    |1  |{[{"a": 1}, {"a": 2}]}|[{1}, {2}]  |
    |1  |{[{"a": 2}, {"a": 4}]}|[{2}, {4}]  |
    |2  |null                  |null        |
    +---+----------------------+------------+
    <BLANKLINE>

    :param spark: the SparkContext
    :param df: a DataFrame
    :param columns: A list of column names
    :return: a new DataFrame
    """
    if type(columns) == str:
        columns = [columns]

    def wrap_json(col: str) -> Column:
        """Transforms a string column into the json string '{"col_name": col_value}'

        It is necessary to wrap the json because the schema inference is incorrect for json that are arrays

        Examples:
        df.show()
        +---+--------------------+
        | id|               json1|
        +---+--------------------+
        |  1|[{"a": 1}, {"a": 2}]|
        |  1|[{"a": 2}, {"a": 4}]|
        +---+--------------------+
        schema = job.spark.read.json(df.select('json1').rdd.map(lambda r: r[0])).schema
        schema.simpleString()
        returns 'struct<a:bigint>' but we the schema we need is 'array<struct<a:bigint>>'

        :param col:
        :return:
        """
        return f.concat(f.lit('{"'), f.lit(col), f.lit('": '), f.coalesce(col, f.lit("null")), f.lit("}"))

    res = df
    for col in columns:
        # >>> res.show()
        # +---+--------------------+
        # | id|               json1|
        # +---+--------------------+
        # |  1|[{"a": 1}, {"a": 2}]|
        # |  1|[{"a": 2}, {"a": 4}]|
        # |  2|                null|
        # +---+--------------------+
        res = res.withColumn(col, wrap_json(col))
        # >>> res.show(10, False)
        # +---+-------------------------------+
        # |id |json1                          |
        # +---+-------------------------------+
        # |1  |{"json1": [{"a": 1}, {"a": 2}]}|
        # |1  |{"json1": [{"a": 2}, {"a": 4}]}|
        # |2  |{"json1": null}                |
        # +---+-------------------------------+
        schema = spark.read.json(res.select(quote(col)).rdd.map(lambda r: r[0])).schema
        # >>> schema.simpleString()
        # 'struct<json1:array<struct<a:bigint>>>'
        res = res.withColumn(col, f.from_json(quote(col), schema).alias(col)[col])
        # >>> res.printSchema()
        # root
        #  |-- id: integer (nullable = true)
        #  |-- json1: array (nullable = true)
        #  |    |-- element: struct (containsNull = true)
        #  |    |    |-- a: long (nullable = true)
        # >>> res.show()
        # +---+----------+
        # | id|     json1|
        # +---+----------+
        # |  1|[[1], [2]]|
        # |  1|[[2], [4]]|
        # |  2|      null|
        # +---+----------+
    return res


def flatten_struct_in_df(struct_df: DataFrame, max_depth: int) -> DataFrame:
    """Flatten a struct column in a spark dataframe.

    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame(
    ...     [
    ...         (1, {"a": 1, "b": {"c": 2, "d": 3}}),
    ...         (2, None),
    ...         (3, {"a": 1, "b": {"c": 2, "d": 3}})
    ...     ],
    ...     "id INT, s STRUCT<a:INT, b:STRUCT<c:INT, d:INT>>"
    ... )
    >>> df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s: struct (nullable = true)
     |    |-- a: integer (nullable = true)
     |    |-- b: struct (nullable = true)
     |    |    |-- c: integer (nullable = true)
     |    |    |-- d: integer (nullable = true)
    <BLANKLINE>
    >>> df.show()
    +---+-----------+
    | id|          s|
    +---+-----------+
    |  1|{1, {2, 3}}|
    |  2|       null|
    |  3|{1, {2, 3}}|
    +---+-----------+
    <BLANKLINE>
    >>> flat_df = flatten_struct_in_df(df, 1)
    >>> flat_df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s_a: integer (nullable = true)
     |-- s_b: struct (nullable = true)
     |    |-- c: integer (nullable = true)
     |    |-- d: integer (nullable = true)
    <BLANKLINE>
    >>> flat_df.show()
    +---+----+------+
    | id| s_a|   s_b|
    +---+----+------+
    |  1|   1|{2, 3}|
    |  2|null|  null|
    |  3|   1|{2, 3}|
    +---+----+------+
    <BLANKLINE>

    :param struct_df: a spark dataframe
    :param max_depth: The maximum depth to flatten recursively (ex: 2 will flatten 1 and 2)
    :return: A spark dataFrame
    """
    flat_cols, struct_cols, flat_df = [], [], []
    flat_cols.append([f.col(c.name) for c in struct_df.schema if not isinstance(c.dataType, StructType)])
    struct_cols.append([c.name for c in struct_df.schema if isinstance(c.dataType, StructType)])
    flat_df.append(
        struct_df.select(
            flat_cols[0]
            + [
                f.col(sc + "." + c).alias(sc + "_" + c)
                for sc in struct_cols[0]
                for c in struct_df.select(sc + ".*").columns
            ]
        )
    )
    for i in range(1, max_depth):
        flat_cols.append([f.col(c.name) for c in flat_df[i - 1].schema if not isinstance(c.dataType, StructType)])
        struct_cols.append([c.name for c in flat_df[i - 1].schema if isinstance(c.dataType, StructType)])
        flat_df.append(
            flat_df[i - 1].select(
                flat_cols[i]
                + [
                    f.col(sc + "." + c).alias(sc + "_" + c)
                    for sc in struct_cols[i]
                    for c in flat_df[i - 1].select(sc + ".*").columns
                ]
            )
        )
    return flat_df[-1]


def flatten(df: DataFrame, separator: str = ".") -> DataFrame:
    """Flattens all the struct columns of a DataFrame

    Nested fields names will be joined together using the specified separator

    Examples:

    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame(
    ...         [(1, {"a": 1, "b": {"c": 1, "d": 1}})],
    ...         "id INT, s STRUCT<a:INT, b:STRUCT<c:INT, d:INT>>"
    ...      )
    >>> df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s: struct (nullable = true)
     |    |-- a: integer (nullable = true)
     |    |-- b: struct (nullable = true)
     |    |    |-- c: integer (nullable = true)
     |    |    |-- d: integer (nullable = true)
    <BLANKLINE>
    >>> flatten(df).printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s.a: integer (nullable = true)
     |-- s.b.c: integer (nullable = true)
     |-- s.b.d: integer (nullable = true)
    <BLANKLINE>
    >>> df = spark.createDataFrame(
    ...         [(1, {"a.a1": 1, "b.b1": {"c.c1": 1, "d.d1": 1}})],
    ...         "id INT, `s.s1` STRUCT<`a.a1`:INT, `b.b1`:STRUCT<`c.c1`:INT, `d.d1`:INT>>"
    ... )
    >>> df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s.s1: struct (nullable = true)
     |    |-- a.a1: integer (nullable = true)
     |    |-- b.b1: struct (nullable = true)
     |    |    |-- c.c1: integer (nullable = true)
     |    |    |-- d.d1: integer (nullable = true)
    <BLANKLINE>
    >>> flatten(df, "?").printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s.s1?a.a1: integer (nullable = true)
     |-- s.s1?b.b1?c.c1: integer (nullable = true)
     |-- s.s1?b.b1?d.d1: integer (nullable = true)
    <BLANKLINE>

    :param df: a DataFrame
    :param separator: It might be useful to change the separator when some DataFrame's column names already contain dots
    :return: a flattened DataFrame
    """
    # The idea is to recursively write a "SELECT s.b.c as `s.b.c`" for each nested column.
    cols = []

    def expand_struct(struct: StructType, col_stack: List[str]) -> None:
        for field in struct:
            if type(field.dataType) == StructType:
                expand_struct(field.dataType, col_stack + [field.name])
            else:
                column = f.col(".".join(quote_columns(col_stack + [field.name])))
                cols.append(column.alias(separator.join(col_stack + [field.name])))

    expand_struct(df.schema, col_stack=[])
    return df.select(cols)


def _build_nested_struct_tree(columns: List[str], struct_separator: str) -> OrderedTree:
    """Given a list of flattened column names and a separator

    >>> _build_nested_struct_tree(["id", "s.a", "s.b.c", "s.b.d"], ".")
    OrderedDict([('id', None), ('s', OrderedDict([('a', None), ('b', OrderedDict([('c', None), ('d', None)]))]))])

    :param columns: Name of the flattened columns
    :param struct_separator: Separator used in the column names for structs
    :return:
    """

    def rec_insert(node: OrderedTree, col: str) -> None:
        if struct_separator in col:
            struct, subcol = col.split(struct_separator, 1)
            if struct not in node:
                node[struct] = OrderedDict()
            rec_insert(node[struct], subcol)
        else:
            node[col] = None

    tree: OrderedTree = OrderedDict()
    for c in columns:
        rec_insert(tree, c)
    return tree


def _build_struct_from_tree(node: OrderedTree, separator: str, prefix: str = "") -> List[Column]:
    """Given an intermediate tree representing a nested struct, build a Spark Column
    that represents this nested structure.

    >>> spark = get_spark_session("doctest")
    >>> tree = OrderedDict([('b!', OrderedDict([
    ...      ('c', None),
    ...      ('d', None)
    ...    ]))])
    >>> _build_struct_from_tree(tree, ".") # noqa: E501
    [Column<'CASE WHEN ((true AND (`b!.c` AS c IS NULL)) AND (`b!.d` AS d IS NULL)) THEN NULL ELSE struct(`b!.c` AS c, `b!.d` AS d) END AS `b!`'>]

    :param node:
    :param separator:
    :param prefix:
    :return:
    """
    cols = []
    for key, value in node.items():
        if value is None:
            cols.append(f.col(quote(prefix + key)).alias(key))
        else:
            fields = _build_struct_from_tree(value, separator, prefix + key + separator)
            # We don't want structs where all fields are null so we check for this
            all_fields_are_null = f.lit(True)
            for field in fields:
                all_fields_are_null = all_fields_are_null & f.isnull(field)

            struct_col = f.when(all_fields_are_null, f.lit(None)).otherwise(f.struct(*fields)).alias(key)
            cols.append(struct_col)
    return cols


def unflatten(df: DataFrame, separator: str = ".") -> DataFrame:
    """Reverse of the flatten operation
    Nested fields names will be separated from each other using the specified separator

    Example:

    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame([(1, 1, 1, 1)], "id INT, `s.a` INT, `s.b.c` INT, `s.b.d` INT")
    >>> df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s.a: integer (nullable = true)
     |-- s.b.c: integer (nullable = true)
     |-- s.b.d: integer (nullable = true)
    <BLANKLINE>
    >>> unflatten(df).printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s: struct (nullable = true)
     |    |-- a: integer (nullable = true)
     |    |-- b: struct (nullable = true)
     |    |    |-- c: integer (nullable = true)
     |    |    |-- d: integer (nullable = true)
    <BLANKLINE>
    >>> df = spark.createDataFrame([(1, 1, 1)], "id INT, `s.s1?a.a1` INT, `s.s1?b.b1` INT")
    >>> df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s.s1?a.a1: integer (nullable = true)
     |-- s.s1?b.b1: integer (nullable = true)
    <BLANKLINE>
    >>> unflatten(df, "?").printSchema()
    root
     |-- id: integer (nullable = true)
     |-- s.s1: struct (nullable = true)
     |    |-- a.a1: integer (nullable = true)
     |    |-- b.b1: integer (nullable = true)
    <BLANKLINE>

    :param df: a DataFrame
    :param separator: It might be useful to change the separator when some DataFrame's column names already contain dots
    :return: a flattened DataFrame
    """
    # The idea is to recursively write a "SELECT struct(a, struct(s.b.c, s.b.d)) as s" for each nested column.
    # There is a little twist as we don't want to rebuild the struct if all its fields are null, so we add a CASE WHEN

    def has_structs(df: DataFrame) -> bool:
        struct_fields = [field for field in df.schema if isinstance(field.dataType, StructType)]
        return len(struct_fields) > 0

    if has_structs(df):
        df = flatten(df)

    tree = _build_nested_struct_tree(df.columns, separator)
    cols = _build_struct_from_tree(tree, separator)
    return df.select(cols)


def flatten_schema(schema: StructType, explode: bool) -> StructType:
    """Transform a DataFrame schema into a new schema where all structs have been flattened.
    The field names are kept, with a '.' separator for struct fields.
    If `explode` option is set, arrays are exploded with a '!' separator.

    Example:
    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame(
    ...     [
    ...         (1, {"a": 1, "b": [{"c": 2, "d": 3}], "e": [4, 5]}),
    ...         (2, None),
    ...     ],
    ...     "id INT, s STRUCT<a:INT, b:ARRAY<STRUCT<c:INT, d:INT, e:ARRAY<INT>>>>"
    ... )
    >>> df.schema.simpleString()
    'struct<id:int,s:struct<a:int,b:array<struct<c:int,d:int,e:array<int>>>>>'
    >>> flatten_schema(df.schema, explode=True).simpleString()
    'struct<id:int,s.a:int,s.b!.c:int,s.b!.d:int,s.b!.e!:int>'
    >>> flatten_schema(df.schema, explode=False).simpleString()
    'struct<id:int,s.a:int,s.b:array<struct<c:int,d:int,e:array<int>>>>'

    :param schema: A Spark DataFrame's Schema
    :param explode: If set, arrays are exploded and a '!' separator is appended to their name.
    :return:
    """

    def flatten_data_type(
        prefix: str, data_type: DataType, is_nullable: bool, metadata: Dict[str, str]
    ) -> List[StructField]:
        if isinstance(data_type, StructType):
            return flatten_struct_type(data_type, is_nullable, prefix + ".")
        elif isinstance(data_type, ArrayType) and explode:
            return flatten_data_type(
                prefix + "!", data_type.elementType, is_nullable or data_type.containsNull, metadata
            )
        else:
            return [StructField(prefix, data_type, is_nullable, metadata)]

    def flatten_struct_type(schema: StructType, previous_nullable: bool = False, prefix: str = "") -> List[StructField]:
        res = []
        for field in schema:
            if isinstance(field.dataType, StructType):
                res += flatten_struct_type(
                    field.dataType, previous_nullable or field.nullable, prefix + field.name + "."
                )
            else:
                res += flatten_data_type(
                    prefix + field.name, field.dataType, previous_nullable or field.nullable, field.metadata
                )
        return res

    return StructType(flatten_struct_type(schema))


def structify(df: DataFrame, struct_name: str, primary_key: Union[str, Iterable[str]], deduplicate: bool) -> DataFrame:
    """Encapsulates all columns of the given dataframe inside a STRUCT.
    If 'deduplicate' is set to true, the dataframe will be grouped by primary_key
    and the STRUCT will be aggregated into an ARRAY<STRUCT>.

    :param df: a Dataframe
    :param struct_name: name of the struct that will encapsulate all the columns
    :param primary_key: name of the columns to keep out of the struct
    :param deduplicate: If set to true, the STRUCT will be aggregated into an ARRAY<STRUCT>.
    :return: a dataframe
    """
    if type(primary_key) == str:
        primary_key = [primary_key]
    other_columns = [col for col in df.columns if col not in primary_key]
    df2 = df.select(*primary_key, f.struct(*other_columns).alias(struct_name))
    if deduplicate:
        return df2.groupBy(*primary_key).agg(f.collect_list(struct_name).alias(struct_name))
    else:
        return df2


def get_sortable_columns(df: DataFrame) -> List[str]:
    """Get the list of all columns that are sortable in a Dataframe.
    For instance, a column of type Map<K, V> is not sortable.

    :param df:
    :return:
    """
    return [quote(type.name) for type in df.schema if not isinstance(type.dataType, MapType)]


def unpivot(df: DataFrame, pivot_columns: List[str], key_alias: str = "key", value_alias: str = "value") -> DataFrame:
    """Unpivot the given DataFrame along the specified pivot columns.
    All columns that are not pivot columns should have the same type.

    Example:

    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame([
    ...    (2018, "Orange",  None, 4000, None),
    ...    (2018, "Beans",   None, 1500, 2000),
    ...    (2018, "Banana",  2000,  400, None),
    ...    (2018, "Carrots", 2000, 1200, None),
    ...    (2019, "Orange",  5000, None, 5000),
    ...    (2019, "Beans",   None, 1500, 2000),
    ...    (2019, "Banana",  None, 1400,  400),
    ...    (2019, "Carrots", None,  200, None),
    ...  ], "year INT, product STRING, Canada INT, China INT, Mexico INT"
    ... )
    >>> df.show()
    +----+-------+------+-----+------+
    |year|product|Canada|China|Mexico|
    +----+-------+------+-----+------+
    |2018| Orange|  null| 4000|  null|
    |2018|  Beans|  null| 1500|  2000|
    |2018| Banana|  2000|  400|  null|
    |2018|Carrots|  2000| 1200|  null|
    |2019| Orange|  5000| null|  5000|
    |2019|  Beans|  null| 1500|  2000|
    |2019| Banana|  null| 1400|   400|
    |2019|Carrots|  null|  200|  null|
    +----+-------+------+-----+------+
    <BLANKLINE>
    >>> unpivot(df, ['year', 'product'], key_alias='country', value_alias='total').show(100)
    +----+-------+-------+-----+
    |year|product|country|total|
    +----+-------+-------+-----+
    |2018| Orange| Canada| null|
    |2018| Orange|  China| 4000|
    |2018| Orange| Mexico| null|
    |2018|  Beans| Canada| null|
    |2018|  Beans|  China| 1500|
    |2018|  Beans| Mexico| 2000|
    |2018| Banana| Canada| 2000|
    |2018| Banana|  China|  400|
    |2018| Banana| Mexico| null|
    |2018|Carrots| Canada| 2000|
    |2018|Carrots|  China| 1200|
    |2018|Carrots| Mexico| null|
    |2019| Orange| Canada| 5000|
    |2019| Orange|  China| null|
    |2019| Orange| Mexico| 5000|
    |2019|  Beans| Canada| null|
    |2019|  Beans|  China| 1500|
    |2019|  Beans| Mexico| 2000|
    |2019| Banana| Canada| null|
    |2019| Banana|  China| 1400|
    |2019| Banana| Mexico|  400|
    |2019|Carrots| Canada| null|
    |2019|Carrots|  China|  200|
    |2019|Carrots| Mexico| null|
    +----+-------+-------+-----+
    <BLANKLINE>

    :param df: a DataFrame
    :param pivot_columns: The list of columns names on which to perform the pivot
    :param key_alias: alias given to the 'key' column
    :param value_alias: alias given to the 'value' column
    :return:
    """
    pivoted_columns = [(c, t) for (c, t) in df.dtypes if c not in pivot_columns]
    cols, types = zip(*pivoted_columns)

    # Check that all columns have the same type.
    assert_true(
        len(set(types)) == 1,
        ("All pivoted columns should be of the same type:\n Pivoted columns are: %s" % pivoted_columns),
    )

    # Create and explode an array of (column_name, column_value) structs
    kvs = f.explode(
        f.array(*[f.struct(f.lit(c).alias(key_alias), f.col(quote(c)).alias(value_alias)) for c in cols])
    ).alias("kvs")

    return df.select([f.col(c) for c in quote_columns(pivot_columns)] + [kvs]).select(
        quote_columns(pivot_columns) + ["kvs.*"]
    )


def union_by_name_sparse(left_df: DataFrame, right_df: DataFrame, use_flatten: bool = False) -> DataFrame:
    """Builds the union of two DataFrame but ignore columns that are from the other DataFrame

    :param left_df:
    :param right_df:
    :param use_flatten: If set to true, each input DataFrame will be flattened and the union unflattened
    :return:
    """
    if use_flatten:
        left_df = flatten(left_df)
        right_df = flatten(right_df)
    left_cols = set(left_df.columns)
    right_cols = set(right_df.columns)
    left_only_cols = left_cols.difference(right_cols)
    right_only_cols = right_cols.difference(left_cols)
    left_df = left_df.drop(*left_only_cols)
    right_df = right_df.drop(*right_only_cols)
    res_df = left_df.unionByName(right_df)
    if use_flatten:
        res_df = unflatten(res_df)
    return res_df


def union_data_frames(left_df: DataFrame, right_df: DataFrame, use_flatten: bool = False) -> DataFrame:
    """Union between two DataFrames, even if schemas are different

    :param left_df:
    :param right_df:
    :param use_flatten: If set to true, each input DataFrame will be flattened and the union unflattened
    :return:
    """

    def order_df_and_add_missing_cols(
        df: DataFrame, columns_order_list: List[str], df_missing_fields: Set[str]
    ) -> DataFrame:
        columns = []
        for name in columns_order_list:
            if name not in df_missing_fields:
                columns.append(f.col(quote(name)))
            else:
                columns.append(f.lit(None).alias(name))
        return df.select(columns)

    def order_and_union_data_frames(
        left_df: DataFrame, right_df: DataFrame, right_only_cols: Set[str], left_only_cols: Set[str]
    ) -> DataFrame:

        missing_columns = [f.lit(None).alias(col) for col in right_only_cols]

        left_df = left_df.select([f.col(c) for c in quote_columns(left_df.columns)] + missing_columns)
        right_df = order_df_and_add_missing_cols(right_df, left_df.columns, left_only_cols)
        return left_df.union(right_df)

    if use_flatten:
        left_df = flatten(left_df)
        right_df = flatten(right_df)

    if left_df is None:
        return right_df
    if right_df is None:
        return left_df
    if left_df.columns == right_df.columns:
        return left_df.union(right_df)

    left_df_col_list = set(left_df.columns)
    right_df_col_list = set(right_df.columns)
    left_only_cols = left_df_col_list.difference(right_df_col_list)
    right_only_cols = right_df_col_list.difference(left_df_col_list)
    res_df = order_and_union_data_frames(left_df, right_df, right_only_cols, left_only_cols)
    if use_flatten:
        res_df = unflatten(res_df)
    return res_df


def make_global_df(*dataframes: DataFrame, use_flatten: bool = False) -> DataFrame:
    """For spark 3.2 and beyond, you won't need this anymore, as unionByName does the same stuff:

    df = df1.unionByName(df2, allowMissingColumns=True)

    Example:

    >>> from pyspark import Row
    >>> spark = get_spark_session("doctest")
    >>> df1 = spark.createDataFrame(data=[
    ...   Row(a=1, b="1"),
    ...   Row(a=2, b="2"),
    ... ], schema="a INT, b STRING")
    >>> df2 = spark.createDataFrame(data=[
    ...   Row(a=3, c=True),
    ...   Row(a=4, c=False),
    ... ], schema="a INT, c BOOLEAN")
    >>> make_global_df(df1, df2, use_flatten=False).show()
    +---+----+-----+
    |  a|   b|    c|
    +---+----+-----+
    |  1|   1| null|
    |  2|   2| null|
    |  3|null| true|
    |  4|null|false|
    +---+----+-----+
    <BLANKLINE>

    :param dataframes: Dataframes to be merged into a big one.
    :param use_flatten: If True, split structs into columns while merging DataFrames
    :return: A big merged dataframe, with null values for df with missing columns.
    """
    global_df = None
    for df in dataframes:
        if global_df:
            global_df = union_data_frames(global_df, df, use_flatten=use_flatten)
        else:
            global_df = df
    return global_df


def coalesce_join(
    left_df: DataFrame, right_df: DataFrame, on: Column, how: str, invert_cols: Optional[List[str]] = None
) -> DataFrame:
    """Perform a join between two DataFrames and coalesces all the fields together,
     using fields from the right side when the ones from the left side are null.
     It can also perform an inverse coalesce on columns indicated in the variable invert_cols

    :param left_df: Left side of the join
    :param right_df: Right side of the join
    :param on: a string for the join column name, a list of column names,
        a join expression (Column), or a list of Columns.
        If `on` is a string or a list of strings indicating the name of the join column(s),
        the column(s) must exist on both sides, and this performs an equi-join.
    :param how: str, default ``inner``. Must be one of: ``inner``, ``cross``, ``outer``,
        ``full``, ``full_outer``, ``left``, ``left_outer``, ``right``, ``right_outer``,
        ``left_semi``, and ``left_anti``.
    :param invert_cols: list of column names on which an inverse coalesce is performed
    :return:
    """
    if invert_cols is None:
        invert_cols = []

    return left_df.join(right_df, on=on, how=how).select(
        [
            f.coalesce(right_df[quote(column)], left_df[quote(column)]).alias(column)
            if column in invert_cols
            else f.coalesce(left_df[quote(column)], right_df[quote(column)]).alias(column)
            for column in left_df.columns
        ]
    )


def sanitize_fields_name(
    spark: SparkSession, df: DataFrame, unwanted_characters: str = '-&"[].', replacement_character: str = "_"
) -> DataFrame:
    """Performs field names sanitization by removing unwanted character
    and adding an '_' if field name begin with a digit.

    Example:
    >>> from pyspark.sql import Row
    >>> spark = get_spark_session("doctest")
    >>> schema = StructType(
    ... [
    ...     StructField('a-b', StringType()),
    ...     StructField('c', StructType(
    ...         [
    ...         StructField('d-e', StringType()),
    ...         StructField('f', StringType())
    ...         ]
    ...         )
    ...     ),
    ...     StructField('g', MapType(StringType(),StringType())),
    ...     StructField('1h', StringType())
    ... ])
    >>> data= [ Row('a-b',('1','2'),{'h&i':'1','j-k':'l-m'},'3')]
    >>> df = spark.createDataFrame(data, schema)
    >>> df.printSchema()
    root
     |-- a-b: string (nullable = true)
     |-- c: struct (nullable = true)
     |    |-- d-e: string (nullable = true)
     |    |-- f: string (nullable = true)
     |-- g: map (nullable = true)
     |    |-- key: string
     |    |-- value: string (valueContainsNull = true)
     |-- 1h: string (nullable = true)
    <BLANKLINE>
    >>> sanitize_fields_name(spark, df).printSchema()
    root
     |-- a_b: string (nullable = true)
     |-- c: struct (nullable = true)
     |    |-- d_e: string (nullable = true)
     |    |-- f: string (nullable = true)
     |-- g: map (nullable = true)
     |    |-- key: string
     |    |-- value: string (valueContainsNull = true)
     |-- _1h: string (nullable = true)
    <BLANKLINE>
    >>> df.show(1,False)
    +---+------+----------------------+---+
    |a-b|c     |g                     |1h |
    +---+------+----------------------+---+
    |a-b|{1, 2}|{j-k -> l-m, h&i -> 1}|3  |
    +---+------+----------------------+---+
    <BLANKLINE>
    >>> sanitize_fields_name(spark, df).show(1,False)
    +---+------+----------------------+---+
    |a_b|c     |g                     |_1h|
    +---+------+----------------------+---+
    |a-b|{1, 2}|{j-k -> l-m, h&i -> 1}|3  |
    +---+------+----------------------+---+
    <BLANKLINE>

    :spark: sparkSession to use
    :df: Dataframe to sanitize
    :unwanted_characters: string composed of each character to be replaced
    :replacement_character: character used to replace unwanted character
    :return: sanitized Dataframe
    """

    # source : https://stackoverflow.com/questions/43004849/rename-nested-field-in-spark-dataframe

    def sanitize_field_name(s: str) -> str:
        new_name = s
        for char in unwanted_characters:
            new_name = new_name.replace(char, replacement_character)
        if new_name[0] in "0123456789":
            new_name = "_" + new_name
        return new_name

    def sanitize_field(field: StructField) -> StructField:
        field = copy(field)
        field.name = sanitize_field_name(field.name)
        field.dataType = clean_schema(field.dataType)
        return field

    def clean_schema(datatype: DataType) -> StructType:
        datatype = copy(datatype)
        # If the type is a StructType we need to recurse otherwise we can return since
        # we've reached the leaf node
        if isinstance(datatype, StructType):
            # We call our sanitizer for all top level fields
            datatype.fields = [sanitize_field(field) for field in datatype.fields]
        elif isinstance(datatype, ArrayType):
            datatype.elementType = clean_schema(datatype.elementType)
        return cast(StructType, datatype)

    return spark.createDataFrame(df.rdd, clean_schema(df.schema))


def ascending_forest_traversal(df: DataFrame, id: str, parent_id: str) -> DataFrame:
    """Given a DataFrame representing a labeled forest with columns "id", "parent_id" and other label columns,
    performs a graph traversal that will return a DataFrame with the same schema that gives for each node
    the labels of it's furthest ancestor.

    This algorithm is optimized for lowly-connected graphs that fit in RAM.
    In other words, for a graph G = (V, E) we assume that |E| << |V|

    It has a security against dependency cycles, but no security preventing
    a combinatorics explosion if some nodes have more than one parent.

    Example:

    >>> spark = get_spark_session("doctest")

    # Given a DataFrame with pokemon attributes and evolution links

    >>> df = spark.sql('''
    ...     SELECT
    ...       col1 as `pokemon.id`,
    ...       col2 as `pokemon.evolve_to_id`,
    ...       col3 as `pokemon.name`,
    ...       col4 as `pokemon.types`
    ...     FROM VALUES
    ...       (4, 5, 'Charmander', ARRAY('Fire')),
    ...       (5, 6, 'Charmeleon', ARRAY('Fire')),
    ...       (6, NULL, 'Charizard', ARRAY('Fire', 'Flying'))
    ... ''')
    >>> df.show()
    +----------+--------------------+------------+--------------+
    |pokemon.id|pokemon.evolve_to_id|pokemon.name| pokemon.types|
    +----------+--------------------+------------+--------------+
    |         4|                   5|  Charmander|        [Fire]|
    |         5|                   6|  Charmeleon|        [Fire]|
    |         6|                null|   Charizard|[Fire, Flying]|
    +----------+--------------------+------------+--------------+
    <BLANKLINE>

    # We compute a DataFrame that for each pokemon.id gives the attributes of its highest level of evolution

    >>> ascending_forest_traversal(df, "pokemon.id", "pokemon.evolve_to_id").orderBy("`pokemon.id`").show()
    +----------+--------------------+------------+--------------+
    |pokemon.id|pokemon.evolve_to_id|pokemon.name| pokemon.types|
    +----------+--------------------+------------+--------------+
    |         4|                null|   Charizard|[Fire, Flying]|
    |         5|                null|   Charizard|[Fire, Flying]|
    |         6|                null|   Charizard|[Fire, Flying]|
    +----------+--------------------+------------+--------------+
    <BLANKLINE>

    :param df: a Spark DataFrame
    :param id: name of the column that represent the node's ids
    :param parent_id: name of the column that represent the parent node's ids
    :return: a DataFrame with the same schema as the input DataFrame that gives
        for each node the labels of it's furthest ancestor
    """
    assert_true(id in df.columns, "Could not find column %s in dataframe's columns: %s" % (id, df.columns))
    assert_true(
        parent_id in df.columns, "Could not find column %s in dataframe's columns: %s" % (parent_id, df.columns)
    )
    df = df.repartition(200, f.col(quote(id))).persist()
    df_null = df.where(f.col(quote(parent_id)).isNull())
    df_not_null = df.where(f.col(quote(parent_id)).isNotNull()).persist()
    do_continue = True

    while do_continue:
        joined_df_not_null = (
            df_not_null.alias("a")
            .join(df.alias("b"), f.col("a." + quote(parent_id)) == f.col("b." + quote(id)), "left")
            .select(
                f.col("a." + quote(id)).alias(id),
                f.when(f.col("b." + quote(parent_id)) == f.col("a." + quote(id)), f.lit(None))
                .otherwise(f.col("b." + quote(parent_id)))
                .alias(parent_id),
                *[
                    f.coalesce(f.col("b." + quote(col)), f.col("a." + quote(col))).alias(col)
                    for col in df.columns
                    if col not in (id, parent_id)
                ],
            )
        )
        joined_df_not_null = joined_df_not_null.persist()
        new_df_not_null = joined_df_not_null.where(f.col(quote(parent_id)).isNotNull()).persist()
        do_continue = new_df_not_null.count() > 0
        new_df_null = df_null.union(joined_df_not_null.where(f.col(quote(parent_id)).isNull()))

        df_not_null = new_df_not_null
        df_null = new_df_null

    return df_null.union(df_not_null)


def empty_array(element_type: Union[DataType, str]) -> Column:
    """Creates an empty Spark array column of the specified type.
    This is a workaround because the Spark method typedLit is not available in PySpark

    :param element_type: the type of the array's element
    :return: a Spark Column representing an empty array.
    """
    return f.array_except(f.array(f.lit(None).cast(element_type)), f.array(f.lit(None)))


def to_generic_struct(*columns: str, col_name_alias: str = "name", col_value_alias: str = "value") -> Column:
    """Spark function that transforms a set of columns into an
    ARRAY<STRUCT<name: STRING, value: STRING> (column_name -> column_value)

    Example:

    >>> spark = get_spark_session("doctest")
    >>> df = spark.sql('''
    ...     SELECT
    ...       col1 as `pokemon.id`,
    ...       col2 as `pokemon.name`,
    ...       col3 as `pokemon.types`
    ...     FROM VALUES
    ...       (4, 'Charmander', ARRAY(NAMED_STRUCT('type', 'Fire'))),
    ...       (5, 'Charmeleon', ARRAY(NAMED_STRUCT('type', 'Fire'))),
    ...       (6, 'Charizard',  ARRAY(NAMED_STRUCT('type', 'Fire'), NAMED_STRUCT('type', 'Flying')))
    ... ''')
    >>> df.show()
    +----------+------------+------------------+
    |pokemon.id|pokemon.name|     pokemon.types|
    +----------+------------+------------------+
    |         4|  Charmander|          [{Fire}]|
    |         5|  Charmeleon|          [{Fire}]|
    |         6|   Charizard|[{Fire}, {Flying}]|
    +----------+------------+------------------+
    <BLANKLINE>
    >>> res = df.select(to_generic_struct("pokemon.id", "pokemon.name", "pokemon.types").alias('values'))
    >>> res.printSchema()
    root
     |-- values: array (nullable = false)
     |    |-- element: struct (containsNull = false)
     |    |    |-- name: string (nullable = false)
     |    |    |-- value: string (nullable = false)
    <BLANKLINE>
    >>> res.show(10, False)
    +---------------------------------------------------------------------------------+
    |values                                                                           |
    +---------------------------------------------------------------------------------+
    |[{pokemon.id, 4}, {pokemon.name, Charmander}, {pokemon.types, [{Fire}]}]         |
    |[{pokemon.id, 5}, {pokemon.name, Charmeleon}, {pokemon.types, [{Fire}]}]         |
    |[{pokemon.id, 6}, {pokemon.name, Charizard}, {pokemon.types, [{Fire}, {Flying}]}]|
    +---------------------------------------------------------------------------------+
    <BLANKLINE>

    :param columns:
    :param col_name_alias: (Default="name") alias of the field containing the column names in the returned struct
    :param col_value_alias: (Default="value") alias of the field containing the column values in the returned struct
    :return:
    """
    return f.array(
        *[
            f.struct(f.lit(c).alias(col_name_alias), f.col(quote(c)).astype(StringType()).alias(col_value_alias))
            for c in columns
        ]
    )


def nullable(col: Column) -> Column:
    """Spark function that makes a DataFrame column nullable.
    This is especially useful for literal which are always non-nullable by default.

    Example:

    >>> spark = get_spark_session("doctest")
    >>> from pyspark.sql import functions as f
    >>> df = spark.sql('''SELECT 1 as a''').withColumn("b", f.lit("2"))
    >>> df.printSchema()
    root
     |-- a: integer (nullable = false)
     |-- b: string (nullable = false)
    <BLANKLINE>
    >>> res = df.withColumn('a', nullable(f.col('a'))).withColumn('b', nullable(f.col('b')))
    >>> res.printSchema()
    root
     |-- a: integer (nullable = true)
     |-- b: string (nullable = true)
    <BLANKLINE>

    :param col:
    :return:
    """
    return f.when(~col.isNull(), col)


__split_col_name_regex = re.compile(
    r"""(                   # Matching group 1 [Everything that comes before the first DOT split]
            (?:                 # Non-matching group
                    [^`.]*          # Everything but BACKQUOTE and DOT zero or more times
                    `               # BACKQUOTE
                    [^`]*           # Everything but BACKQUOTE zero or more times [DOTS are allowed between BACKQUOTES]
                    `               # BACKQUOTE
                |               # OR
                    [^`.]+          # Everything but BACKQUOTE and DOT one or more times
            )+                  # Repeat this group one or more times
        )
        ([.]|$)             # Matching group 2 [The separator: DOT or end of line (to match strings without dots)]
    """,
    re.VERBOSE,  # Allows to add comments inside the regex
)

re.compile(r"([^.]+)([.]|$)").findall("a.b.c")


def split_col_name(col_name: str) -> List[str]:
    """Splits a Spark column name representing a nested field into a list of path parts.

    Examples:
    In this example: `a` is a struct containing a field `b`.

    >>> split_col_name("a.b")
    ['a', 'b']
    >>> split_col_name("ab")
    ['ab']

    Field names can contain dots when escaped between backquotes (`)

    >>> split_col_name("`a.b`")
    ['a.b']
    >>> split_col_name("`a.b`.`c.d`")
    ['a.b', 'c.d']
    >>> split_col_name("`a.b`.c.`d`")
    ['a.b', 'c', 'd']

    Field names can even contain backquotes, by escaping backquotes with another backquote.

    >>> split_col_name("`a``b`")
    ['a`b']
    >>> split_col_name("`.a.``.b.`")
    ['.a.`.b.']
    >>> split_col_name("`ab`.`c``d`.fg")
    ['ab', 'c`d', 'fg']

    :param col_name:
    :return:
    """
    col_parts = [unquote(match[0]) for match in __split_col_name_regex.findall(col_name)]
    return col_parts


def get_col_type(schema: StructType, col_name: str) -> DataType:
    """Fetch recursively the DataType of a column inside a DataFrame schema (or more generally any StructType)

    Example:
    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame([
    ...      (1, {"a.b" : {"c.d": 1, "e": "1", "f`g": True}}),
    ...      (2, {"a.b" : {"c.d": 2, "e": "2", "f`g": True}}),
    ...      (3, {"a.b" : {"c.d": 3, "e": "3", "f`g": True}}),
    ...   ],
    ...   "id INT, `a.b` STRUCT<`c.d`:INT, e:STRING, `f``g`:BOOLEAN>"
    ... )
    >>> get_col_type(df.schema, "`a.b`.`c.d`").simpleString()
    'int'
    >>> get_col_type(df.schema, "`a.b`.e").simpleString()
    'string'
    >>> get_col_type(df.schema, "`a.b`.`f``g`").simpleString()
    'boolean'

    :param schema: the DataFrame schema (or StructField) in which the column type will be fetched.
    :param col_name: the name of the column to get
    :return:
    """
    col_parts = split_col_name(col_name)

    def get_col(col: str, fields: List[StructField]) -> StructField:
        for field in fields:
            if field.name == col:
                return field
        raise ValueError(f'Cannot resolve column name "{col_name}"')

    struct: Union[StructType, DataType] = schema
    for col_part in col_parts:
        assert_true(isinstance(struct, StructType))
        struct = cast(StructType, struct)
        struct = get_col(col_part, struct.fields).dataType
    assert_true(isinstance(struct, DataType))
    return cast(DataType, struct)


def to_generic_typed_struct(df: DataFrame, col_names: List[str]) -> DataFrame:
    """Transforms the specified struct columns of a Dataframe into generic typed struct columns with the following
    generic schema (based on https://spark.apache.org/docs/2.4.5/sql-reference.html#data-types):

    .. code-block::

        STRUCT<
            key: STRING, -- (name of the field inside the struct)
            type: STRING, -- (type of the field inside the struct)
            value: STRUCT< -- (all the fields will be null except for the one with the correct type)
                date: DATE,
                timestamp: TIMESTAMP,
                int: LONG,
                float: DOUBLE,
                boolean: BOOLEAN,
                string: STRING,
                bytes: BINARY
            >
        >

    The following spark types will be automatically cast into the more generic following types:

    - tinyint, smallint, int -> bigint
    - float, decimal -> double

    Limitations
    -----------
    Currently, complex field types (structs, maps, arrays) are not supported.
    All fields of the struct columns to convert must be of basic types.

    Examples
    --------

    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame(
    ...     [(1, {"first.name": "Jacques", "age": 25, "is.an.adult": True}),
    ...      (2, {"first.name": "Michel", "age": 12, "is.an.adult": False}),
    ...      (3, {"first.name": "Marie", "age": 36, "is.an.adult": True})],
    ...     "id INT, `person.struct` STRUCT<`first.name`:STRING, age:INT, `is.an.adult`:BOOLEAN>"
    ... )
    >>> df.show(truncate=False)
    +---+-------------------+
    |id |person.struct      |
    +---+-------------------+
    |1  |{Jacques, 25, true}|
    |2  |{Michel, 12, false}|
    |3  |{Marie, 36, true}  |
    +---+-------------------+
    <BLANKLINE>
    >>> res = to_generic_typed_struct(df, ["`person.struct`"])
    >>> res.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- person.struct: array (nullable = false)
     |    |-- element: struct (containsNull = false)
     |    |    |-- key: string (nullable = false)
     |    |    |-- type: string (nullable = false)
     |    |    |-- value: struct (nullable = false)
     |    |    |    |-- boolean: boolean (nullable = true)
     |    |    |    |-- bytes: binary (nullable = true)
     |    |    |    |-- date: date (nullable = true)
     |    |    |    |-- float: double (nullable = true)
     |    |    |    |-- int: long (nullable = true)
     |    |    |    |-- string: string (nullable = true)
     |    |    |    |-- timestamp: timestamp (nullable = true)
    <BLANKLINE>
    >>> res.show(10, False) # noqa: E501
    +---+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    |id |person.struct                                                                                                                                                                                  |
    +---+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    |1  |[{first.name, string, {null, null, null, null, null, Jacques, null}}, {age, int, {null, null, null, null, 25, null, null}}, {is.an.adult, boolean, {true, null, null, null, null, null, null}}]|
    |2  |[{first.name, string, {null, null, null, null, null, Michel, null}}, {age, int, {null, null, null, null, 12, null, null}}, {is.an.adult, boolean, {false, null, null, null, null, null, null}}]|
    |3  |[{first.name, string, {null, null, null, null, null, Marie, null}}, {age, int, {null, null, null, null, 36, null, null}}, {is.an.adult, boolean, {true, null, null, null, null, null, null}}]  |
    +---+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    <BLANKLINE>

    :param df: Dataframe to transform
    :param col_names: List of column names to transform
    :return: a Dataframe with fields transformed into struct with Data type name
    """

    source_to_cast = {
        "date": "date",
        "timestamp": "timestamp",
        "tinyint": "bigint",
        "smallint": "bigint",
        "int": "bigint",
        "bigint": "bigint",
        "float": "double",
        "double": "double",
        "boolean": "boolean",
        "string": "string",
        "binary": "binary",
    }
    """Mapping indicating for each source Spark DataTypes the type into which it will be cast."""

    cast_to_name = {
        "binary": "bytes",
        "bigint": "int",
        "double": "float",
    }
    """Mapping indicating for each already cast Spark DataTypes the name of the corresponding field.
    When missing, the same name will be kept."""

    name_cast = {cast_to_name.get(value, value): value for value in source_to_cast.values()}
    # We make sure the types are sorted
    name_cast = {k: v for k, v in sorted(name_cast.items())}

    def match_regex_types(source_type: str) -> Optional[str]:
        """Matches the source types against regexes to identify more complex types (like Decimal(x, y))"""
        regex_to_cast_types = [(re.compile("decimal(.*)"), "float")]
        for regex, cast_type in regex_to_cast_types:
            if regex.match(source_type) is not None:
                return cast_type
        return None

    def field_to_col(field: StructField, column_name: str) -> Optional[Column]:
        """Transforms the specified field into a generic column"""
        source_type = field.dataType.simpleString()
        cast_type = source_to_cast.get(source_type)
        field_name = column_name + "." + quote(field.name)
        if cast_type is None:
            cast_type = match_regex_types(source_type)
        if cast_type is None:
            print(
                "WARNING: The field {field_name} is of type {source_type} which is currently unsupported. "
                "This field will be dropped.".format(field_name=field_name, source_type=source_type)
            )
            return None
        name_type = cast_to_name.get(cast_type, cast_type)
        return f.struct(
            f.lit(field.name).alias("key"),
            f.lit(name_type).alias("type"),
            # In the code below, we use f.expr instead of f.col because it looks like f.col
            # does not support column names with backquotes in them, but f.expr does :-p
            f.struct(
                *[
                    (f.expr(field_name) if name_type == name_t else f.lit(None)).astype(cast_t).alias(name_t)
                    for name_t, cast_t in name_cast.items()
                ]
            ).alias("value"),
        )

    for col_name in col_names:
        schema = get_col_type(df.schema, col_name)
        assert_true(isinstance(schema, StructType))
        columns = [field_to_col(field, col_name) for field in schema.fields]
        columns_2 = [col for col in columns if col is not None]
        df = df.withColumn(unquote(col_name), f.array(*columns_2).alias("values"))
    return df


def find_wider_type_for_two(t1: DataType, t2: DataType) -> Optional[str]:
    """Python wrapper for Spark's TypeCoercion.find_wider_type_for_two:

    Looking for a widened data type of two given data types with some acceptable loss of precision.
    E.g. there is no common type for double and decimal because double's range
    is larger than decimal, and yet decimal is more precise than double, but in
    union we would cast the decimal into double.

    WARNINGS:

    - the result is a simpleString
    - A SparkSession must already be instantiated for this method to work

    >>> from pyspark.sql.types import DecimalType, LongType, DoubleType, IntegerType
    >>> spark = get_spark_session("doctest")
    >>> find_wider_type_for_two(DecimalType(15, 5), DecimalType(15, 6))
    'decimal(16,6)'
    >>> find_wider_type_for_two(DecimalType(15, 5), DoubleType())
    'double'
    >>> find_wider_type_for_two(LongType(), IntegerType())
    'bigint'
    >>> find_wider_type_for_two(ArrayType(IntegerType()), IntegerType())

    :param t1:
    :param t2:
    :return:
    """
    from pyspark.sql import SparkSession

    spark = SparkSession._instantiatedSession
    sc = spark.sparkContext

    def _to_java_type(t: DataType):
        return spark._jsparkSession.parseDataType(t.json())

    jt1 = _to_java_type(t1)
    jt2 = _to_java_type(t2)
    j_type_coercion = getattr(getattr(sc._jvm.org.apache.spark.sql.catalyst.analysis, "TypeCoercion$"), "MODULE$")
    wider_type = j_type_coercion.findWiderTypeForTwo(jt1, jt2)
    if wider_type.nonEmpty():
        return wider_type.get().simpleString()
    else:
        return None


def get_common_columns(left_schema: StructType, right_schema: StructType) -> List[Column]:
    """Return a list of common Columns between two DataFrame schemas, casting them into the widest common type
    when required.

    When columns have incompatible types, they are simply not cast.

    >>> spark = get_spark_session("doctest")
    >>> df1 = spark.sql('''SELECT 'A' as id, CAST(1 as BIGINT) as d, 'a' as a''')
    >>> df2 = spark.sql('''SELECT 'A' as id, CAST(1 as DOUBLE) as d, ARRAY('a') as a''')
    >>> common_cols = get_common_columns(df1.schema, df2.schema)
    >>> df1.select(common_cols).printSchema()
    root
     |-- id: string (nullable = false)
     |-- d: double (nullable = false)
     |-- a: string (nullable = false)
    <BLANKLINE>
    >>> df2.select(common_cols).printSchema()
    root
     |-- id: string (nullable = false)
     |-- d: double (nullable = false)
     |-- a: array (nullable = false)
     |    |-- element: string (containsNull = false)
    <BLANKLINE>

    :param left_schema:
    :param right_schema:
    :return:
    """
    left_fields = {field.name: field for field in left_schema.fields}
    right_fields = {field.name: field for field in right_schema.fields}

    def get_columns() -> Generator[Column, None, None]:
        for name, left_field in left_fields.items():
            if name in right_fields:
                right_field: StructField = right_fields[name]
                if right_field.dataType == left_field.dataType:
                    common_type = None
                else:
                    common_type = find_wider_type_for_two(left_field.dataType, right_field.dataType)
                if common_type is not None:
                    yield f.col(quote(name)).astype(common_type)
                else:
                    yield f.col(quote(name))

    return list(get_columns())


def multi_join_with_common_key(
    dataframes: List[Tuple[DataFrame, str]], common_keys: Iterable[str], priority_list: List[str]
) -> DataFrame:
    """
    given a list of (dataframe, join_type) and a Tuple of common keys join all those dataframes
    >>> spark = get_spark_session("doctest")
    >>> df1 = spark.sql('''SELECT 'A' as id, ARRAY('a') as a''').cache()
    >>> df2 = spark.sql('''SELECT 'A' as id, CAST(1 as BIGINT) as b''').cache()
    >>> df3 = spark.sql('''SELECT 'A' as id, CAST(5 as DOUBLE) as c''').cache()
    >>> dfs = [(df1, "left_outer"), (df2, "full_outer"), (df3, "left_outer")]
    >>> join_priority_list = ["full_outer", "left_outer"]
    >>> result = multi_join_with_common_key(dfs, ('id',), join_priority_list)
    >>> result.show(truncate=False)
    +---+---+---+---+
    |id |a  |b  |c  |
    +---+---+---+---+
    |A  |[a]|1  |5.0|
    +---+---+---+---+
    <BLANKLINE>
    """
    common_keys = list(common_keys)

    dataframes.sort(key=lambda x: priority_list.index(x[1]))

    res = dataframes[0][0]
    for df, join_type in dataframes[1:]:
        res = res.join(df, common_keys, join_type)

    # But here, we want to have the structs sorted in alphanumeric ordering so we sort them accordingly
    columns = res.drop(*common_keys).columns
    columns.sort()

    return res.select(*common_keys, *columns)


def concat_all_columns(df: DataFrame, result_col_name: str, separator: str) -> DataFrame:
    """
    given a dataframe this function concat all columns and return a dataframe with all columns concatened

    >>> spark = get_spark_session("doctest")
    >>> df1 = spark.sql('''SELECT 'A' as id, 'a' as a''').cache()
    >>> result = concat_all_columns(df1, 'tt', ',')
    >>> result.show(truncate=False)
    +---+
    |tt |
    +---+
    |A,a|
    +---+
    <BLANKLINE>
    """
    return df.select(f.expr("concat_ws('%s',*)" % separator).alias(result_col_name))


def camel_case_fields_to_snake_case(df: DataFrame) -> DataFrame:
    """Given a dataframe this function concat all columns and return a dataframe with all columns concatened

    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame(
    ...        [(1, 'a', [1, 2], {'a': 1, 'b': 2}, (1, [1, 2], {'a': 1, 'b': 2}))],
    ...        "intCol INT, StringCol STRING, arrayCol ARRAY<INT>, MapCol MAP<STRING, INT>, "
    ...        "structCol STRUCT<a: INT, InnerArrayCol: ARRAY<INT>, InnerMapCol: MAP<STRING, INT>>"
    ...    )
    >>> df.show(1, False)
    +------+---------+--------+----------------+-----------------------------+
    |intCol|StringCol|arrayCol|MapCol          |structCol                    |
    +------+---------+--------+----------------+-----------------------------+
    |1     |a        |[1, 2]  |{a -> 1, b -> 2}|{1, [1, 2], {a -> 1, b -> 2}}|
    +------+---------+--------+----------------+-----------------------------+
    <BLANKLINE>
    >>> df.printSchema()
    root
     |-- intCol: integer (nullable = true)
     |-- StringCol: string (nullable = true)
     |-- arrayCol: array (nullable = true)
     |    |-- element: integer (containsNull = true)
     |-- MapCol: map (nullable = true)
     |    |-- key: string
     |    |-- value: integer (valueContainsNull = true)
     |-- structCol: struct (nullable = true)
     |    |-- a: integer (nullable = true)
     |    |-- InnerArrayCol: array (nullable = true)
     |    |    |-- element: integer (containsNull = true)
     |    |-- InnerMapCol: map (nullable = true)
     |    |    |-- key: string
     |    |    |-- value: integer (valueContainsNull = true)
    <BLANKLINE>
    >>> result = camel_case_fields_to_snake_case(df)
    >>> result.show(1, False)
    +-------+----------+---------+----------------+-----------------------------+
    |int_col|string_col|array_col|map_col         |struct_col                   |
    +-------+----------+---------+----------------+-----------------------------+
    |1      |a         |[1, 2]   |{a -> 1, b -> 2}|{1, [1, 2], {a -> 1, b -> 2}}|
    +-------+----------+---------+----------------+-----------------------------+
    <BLANKLINE>
    >>> result.printSchema()
    root
     |-- int_col: integer (nullable = true)
     |-- string_col: string (nullable = true)
     |-- array_col: array (nullable = true)
     |    |-- element: integer (containsNull = true)
     |-- map_col: map (nullable = true)
     |    |-- key: string
     |    |-- value: integer (valueContainsNull = true)
     |-- struct_col: struct (nullable = true)
     |    |-- a: integer (nullable = true)
     |    |-- inner_array_col: array (nullable = true)
     |    |    |-- element: integer (containsNull = true)
     |    |-- inner_map_col: map (nullable = true)
     |    |    |-- key: string
     |    |    |-- value: integer (valueContainsNull = true)
    <BLANKLINE>
    """
    import re

    df = flatten(df)

    def camel_case_to_snake_case(s: str) -> str:
        return re.sub(r"(?<!^)(?<!\.)(?=[A-Z])", "_", s).lower()

    for field in df.schema:
        if isinstance(field.dataType, ArrayType) or isinstance(field.dataType, MapType):
            complex_field_mapping_str = field.simpleString()
            (field_name, field_mapping) = complex_field_mapping_str.split(":", 1)
            df = (
                df.selectExpr(
                    "*", f"CAST(`{field_name}` AS {camel_case_to_snake_case(field_mapping)}) AS tmp_snake_field"
                )
                .withColumn(field_name, f.col("tmp_snake_field"))
                .drop("tmp_snake_field")
            )
        df = df.withColumnRenamed(field.name, camel_case_to_snake_case(field.name))
    df = unflatten(df)
    return df


def remove_columns(df: DataFrame, struct_fields: List[str]) -> DataFrame:
    """Given a dataframe this function remove all columns (or nested element) given in argument with absolute
    name path (based on flatten function)

    >>> spark = get_spark_session("doctest")
    >>> df = spark.createDataFrame(
    ...        [(1, 'a', [1, 2], {'a': 1, 'b': 2}, (1, [1, 2], {'a': 1, 'b': 2}, (1, 2)))],
    ...        "id INT, col1 STRING, col2 ARRAY<INT>, col3 MAP<STRING, INT>, "
    ...        "col4 STRUCT<sub_col1: INT, sub_col2: ARRAY<INT>, sub_col3: MAP<STRING, INT>,"
    ...        "sub_col4: STRUCT<sub_col5 :INT, sub_col6 :INT>>"
    ...    )
    >>> df.show(1, False)
    +---+----+------+----------------+-------------------------------------+
    |id |col1|col2  |col3            |col4                                 |
    +---+----+------+----------------+-------------------------------------+
    |1  |a   |[1, 2]|{a -> 1, b -> 2}|{1, [1, 2], {a -> 1, b -> 2}, {1, 2}}|
    +---+----+------+----------------+-------------------------------------+
    <BLANKLINE>
    >>> df.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- col1: string (nullable = true)
     |-- col2: array (nullable = true)
     |    |-- element: integer (containsNull = true)
     |-- col3: map (nullable = true)
     |    |-- key: string
     |    |-- value: integer (valueContainsNull = true)
     |-- col4: struct (nullable = true)
     |    |-- sub_col1: integer (nullable = true)
     |    |-- sub_col2: array (nullable = true)
     |    |    |-- element: integer (containsNull = true)
     |    |-- sub_col3: map (nullable = true)
     |    |    |-- key: string
     |    |    |-- value: integer (valueContainsNull = true)
     |    |-- sub_col4: struct (nullable = true)
     |    |    |-- sub_col5: integer (nullable = true)
     |    |    |-- sub_col6: integer (nullable = true)
    <BLANKLINE>
    >>> result = remove_columns(df, ['col1', 'col4.sub_col2', 'col4.sub_col4.sub_col6'])
    >>> result.show(1, False)
    +---+------+----------------+--------------------------+
    |id |col2  |col3            |col4                      |
    +---+------+----------------+--------------------------+
    |1  |[1, 2]|{a -> 1, b -> 2}|{1, {a -> 1, b -> 2}, {1}}|
    +---+------+----------------+--------------------------+
    <BLANKLINE>
    >>> result.printSchema()
    root
     |-- id: integer (nullable = true)
     |-- col2: array (nullable = true)
     |    |-- element: integer (containsNull = true)
     |-- col3: map (nullable = true)
     |    |-- key: string
     |    |-- value: integer (valueContainsNull = true)
     |-- col4: struct (nullable = true)
     |    |-- sub_col1: integer (nullable = true)
     |    |-- sub_col3: map (nullable = true)
     |    |    |-- key: string
     |    |    |-- value: integer (valueContainsNull = true)
     |    |-- sub_col4: struct (nullable = true)
     |    |    |-- sub_col5: integer (nullable = true)
    <BLANKLINE>
    """
    for struct_field in struct_fields:
        if "." in struct_field:
            col = struct_field.split(".")[0]
            struct = ".".join(struct_field.split(".")[1:])
            df = df.withColumn(col, f.col(col).dropFields(struct))
        else:
            df = df.drop(*struct_fields)
    return df


def left_join_dataframes_on_nullable_key(left_df: DataFrame, key: str, *dataframes: DataFrame) -> DataFrame:
    """
    This function avoid skews when joining datasets on a nullable key.
    """
    left_with_null_key = left_df.where("%s is null" % key)
    left_with_not_null_key = left_df.where("%s is not null" % key)

    left_with_right_fields = left_with_not_null_key
    for df in dataframes:
        left_with_right_fields = left_with_right_fields.join(df, key, "left")

    return union_data_frames(left_with_right_fields, left_with_null_key)


def convert_column_type(df: DataFrame, input_type: str, converted_type: str) -> DataFrame:
    """
    Cast all the columns of type input_type to converted_type of a dataframe
    """
    from pyspark.sql.functions import col

    for column in df.dtypes:
        col_name, col_type = column[0], column[1]
        if col_type == input_type:
            df = df.withColumn(col_name, col(col_name).cast(converted_type))
    return df
